import numpy as np
import re
from random import randint
from itertools import combinations
from scipy.cluster import hierarchy
from .fasta_gtf import read_fasta_gtf
from .genbank_parse import read_genbank
from .genbank_write import write_genbank
from .dataclass import GenericFeature


score = {
    'gap': -1,
    'feature_intergenic': -1,
    'different_feature': -2,
    'same_feature_same_strand': +2,
    'same_feature_opposite_strand': -5,
    'common_tag': +100
}

score_scalar = 0.01


def _tag_feature(feature, qualifiers, skip='orf[0-9]+'):
    """
    Use all of the words in the <qualifiers> to tag the feature

    Args:
        feature: GenericFeature object

        qualifiers: str, or list of str

        skip: str, regex pattern
            Attribute words matching this pattern will be skipped
            Usually we want to skip very generic patterns like 'orf#####'
    """
    for key, val in feature.attributes:
        if key in qualifiers:
            for word in re.split('[|]|[;]|[,]', str(val)):
                if not re.search(skip, word):
                    feature.tags.append(word.strip())
    # Remove duplicates and make the same combination of words always identical
    feature.tags = list(set(feature.tags))


def _has_common_tag(tags1, tags2, skip='orf[0-9]+'):
    """
    Args:
        tags1: list of str

        tags2: list of str

        skip: str, regex pattern
            Tags matching this pattern will be skipped
            Usually we want to skip very generic patterns like 'orf#####'

    Returns: bool
    """
    for t in tags1:
        if t in tags2:
            if not re.search(skip, t):
                return True
    return False


def _make_color_dict(chromosome_dict):
    """
    From all Feature objects of all Chromosome objects,
        use tags to build a color dictionary, which has the following data structure:
        {
            tuple(feature.tags): color, ...
        {

        in which color is str and has the format: #ffffff (white)

    Args:
        chromosome_dict: dict
        {
            chromosome1.seqname: chromosome1,
            chromosome2.seqname: chromosome2, ...
        }

    Returns: dict
    """
    color_dict = {}
    for chromosome in chromosome_dict.values():
        for feature in chromosome.feature_array:
            # no tag -> skip
            if len(feature.tags) == 0: continue
            # feature.tags is list, convert to tuple for hashing
            key = tuple(feature.tags)
            color_dict.setdefault(key, f'#{randint(0, 256**3):06x}')
    return color_dict


def _set_feature_color(chromosome_dict, color_dict):
    """
    For each feature, use its tag to get a color from color_dict
        and set the color attribute with key='Color', val=color

    For example, red color = #ff0000

    Args:
        chromosome_dict: dict

        color_dict: dict
    """
    for chromosome in chromosome_dict.values():
        for feature in chromosome.feature_array:
            if len(feature.tags) == 0:
                # no tag -> default white color
                color = '#ffffff'
            else:
                # Get color with tag
                color = color_dict[tuple(feature.tags)]
            # Genieous format: /Color="#ff0000"
            # Old color (if any) will be replaced
            feature.set_attribute(key='Color', val=color)


def _compare(feature1, feature2):
    """
    Args:
        feature1: GenericFeature object

        feature2: GenericFeature object

    Returns: float
    """
    X, Y = feature1, feature2

    if X is None:
        return score['gap']

    if Y is None:
        return score['gap']

    if X.type == Y.type == 'intergenic':
        return 0

    if X.type == 'intergenic' or Y.type == 'intergenic':
        return score['feature_intergenic']

    if X.type != Y.type:
        return score['different_feature']

    else: # X.type == Y.type but not 'intergenic'

        if X.strand == Y.strand:

            if _has_common_tag(X.tags, Y.tags):
                return score['common_tag']
            else:
                return score['same_feature_same_strand']

        else: # X.strand != Y.strand
            return score['same_feature_opposite_strand']


def _align(feature_array_1, feature_array_2):
    """
    Aligns two FeatureArray objects and return the optimal score of alignment

    Args:
        feature_array_1: FeatureArray object

        feature_array_2: FeatureArray object

    Returns: float
    """
    L1 = feature_array_1
    L2 = feature_array_2

    M = np.zeros((len(L1)+1, len(L2)+1), dtype=np.int)

    for r in range(len(L1)):
        M[r+1, 0] = M[r, 0] + score['gap']

    for c in range(len(L2)):
        M[0, c+1] = M[0, c] + score['gap']

    for r in range(len(L1)):
        for c in range(len(L2)):
            down_right = M[r, c] + _compare(L1[r], L2[c])
            right = M[r+1, c] + _compare(None, L2[c])
            down = M[r, c+1] + _compare(L1[r], None)

            M[r+1, c+1] = max(down_right, right, down)

    return M[-1, -1]


def _insert_intergenic(feature_array, seqname):
    """
    Args:
        feature_array: FeatureArray object

        seqname: str
    """
    L = []
    last_pos = 0
    for f in feature_array:
        if f.start > last_pos + 1:
            L.append(
                GenericFeature(
                    seqname=seqname,
                    type_='intergenic',
                    start=last_pos + 1,
                    end=f.start - 1,
                    strand='+',
                )
            )
            L.append(f)
        else:  # feature.start <= last_pos + 1
            L.append(f)

        last_pos = f.end
    return L


def get_qualifier(genbank, keywords):
    """
    From a <genbank> file, get the qualifiers that contain <keywords>

    Args:
        genbank: str, path-like
            The input genbank file

        keywords: str, or list of str

    Returns: list of str
    """
    if type(keywords) is str:
        keywords = [keywords]

    qualifiers = {}
    for chromosome in read_genbank(genbank):
        for feature in chromosome.feature_array:
            for qualifier, content in feature.attributes:
                # No need to look for keywords if content is not str
                if type(content) is not str: continue
                for word in keywords:
                    if word in content:
                        qualifiers.setdefault(qualifier, True)
    return list(qualifiers.keys())


def synteny(genbank=None, fasta=None, gtf=None, output=None, ax=None, qualifiers=('name', 'gene', 'label')):
    """
    Args:
        genbank: str, path-like

        fasta: str, path-like
            fasta and gtf together overrides genbank

        gtf: str, path-like

        output: str, path-like
            The output genbank file
            Has to be genbank file because the order of chromosomes is
                the only thing that matters for graphic display

        ax: matplotlib Axes instance

        qualifiers: str or list of str
            Qualifiers (e.g. gene, db_xref) of the genbank feature to be used
                to tag and align features

    Returns: list of str
        The order of seqname after synteny alignment and clustering
    """
    if fasta is not None and gtf is not None:
        chromosome_dict = read_fasta_gtf(fasta, gtf, as_dict=True, circular=False)
    else:
        chromosome_dict = read_genbank(genbank, as_dict=True)

    # For each feature, tag it with its own attributes using the <qualifiers>
    if type(qualifiers) is str:
        qualifiers = [qualifiers]
    for seqname, chromosome in chromosome_dict.items():
        for feature in chromosome.feature_array:
            _tag_feature(feature, qualifiers)

    # Generate a color dictionary from all features of all chromosomes
    # Each unique tag or combination of tags gets a unique color
    # Tags (keys) are hashed by tuple, colors are represented by hexadecimal strings, e.g.
    # {
    #   ('zinc finger', 'DNA binding'): #0f0f0f
    # }
    color_dict = _make_color_dict(chromosome_dict)

    # Set the color of each feature using the color_dict
    _set_feature_color(chromosome_dict, color_dict)

    # simply get a list of seqnames
    seqname_list = list(chromosome_dict.keys())

    # scores is in the format of condensed distance matrix
    #   which is an 1-D array of seqname pairs following the order given by combinations()
    scores = []
    for seqname1, seqname2 in combinations(seqname_list, 2):

        arr1 = chromosome_dict[seqname1].feature_array
        arr2 = chromosome_dict[seqname2].feature_array

        arr1 = _insert_intergenic(arr1, seqname1)
        arr2 = _insert_intergenic(arr2, seqname2)

        scores.append(_align(arr1, arr2))  # compute alignment score

    # Condensed distance matrix, which is literally a list of distance (must be positive)
    # distance = e^(-score), the higher the score, the shorter the distance
    dist_mat = np.exp(
        -np.array(scores) * score_scalar
    )

    # Compute linkage matrix
    linkage_matrix = hierarchy.linkage(dist_mat, 'average')

    # Plot dendrogram and get id order
    id_order = hierarchy.dendrogram(
        Z=linkage_matrix,
        ax=ax,
        orientation='top' # root on top
    )['ivl'] # 'ivl': a list of labels corresponding to the leaf nodes

    # Get actual seqname order with id order
    seqname_order = [seqname_list[int(i)] for i in id_order]

    # Set the contig labels on the x axis
    if ax is not None:
        ax.set_xticklabels(seqname_order, rotation=90, horizontalalignment='right')

    # Re-order Chromosome objects in a list
    data = [chromosome_dict[seqname] for seqname in seqname_order]

    if output is None: output = 'synteny.gb'
    write_genbank(data=data, file=output)

    return seqname_order



import os
import random
from .lowlevel import call, gzip


def _sample_fq(file: str, fraction: float, output: str):
    """
    Randomly subsample the input fastq file

    Args:
        file: path-like
            The input fastq file

        fraction:
            Fraction of reads or sequences to be retrieved

        output: path-like
            The output file name
    """
    fh_in = open(file, 'r')
    fh_out = open(output, 'w')

    # Count the total number of lines in the fastq file
    i = 0
    while fh_in.readline() != '':
        i += 1

    # Go back to the beginning of the file, because it needs to go through the file again
    fh_in.seek(0)

    N = int(i / 4)  # The total count of reads
    n = int(N * fraction)  # The count of reads to be retrieved

    # A random array of int corresponding to which reads to be retrieved
    rand_array = random.sample(list(range(N)), n)
    rand_array.sort()

    # Prepare the iterating variable pos (current line position) and next_pos
    pos = -4
    # 4 lines for each read in fastq, so if the next read is 10th read,
    #     then it starts at 40th line
    next_pos = rand_array.pop(0) * 4
    while len(rand_array) > 0:
        # Read 4 lines at a time
        four_lines = [fh_in.readline() for _ in range(4)]
        # Every time read 4 lines, update the current position pos
        pos += 4
        if pos == next_pos:
            # If pos is at the next position, write four lines
            for line in four_lines:
                fh_out.write(line)
            # Update the next position
            next_pos = rand_array.pop(0) * 4
        # If pos is not at the next position, then just continue onto the next read

    fh_in.close()
    fh_out.close()


def _sample_fa(file: str, fraction: float, output: str):
    """
    Randomly subsample the input fasta file

    Args:
        file: path-like
            The input fasta file

        fraction:
            Fraction of reads or sequences to be retrieved

        output: path-like
            The output file name
    """
    fh_in = open(file, 'r')
    fh_out = open(output, 'w')

    # All line positions of the header line '>'
    all_entry_pos = []
    for i, line in enumerate(fh_in):
        if line.startswith('>'):
            all_entry_pos.append(i)

    # Go back to the beginning of the file, because it needs to go through the file again
    fh_in.seek(0)

    N = len(all_entry_pos)  # Total number of reads
    n = int(N * fraction)  # The number of reads to be retrieved
    # Randomly select a subset of entry positions
    rand_entry_pos = random.sample(all_entry_pos, n)
    rand_entry_pos.sort()

    # Prepare the iterating variable line, i and next_pos
    line = fh_in.readline()
    pos = 0
    next_pos = rand_entry_pos.pop(0)
    while True:
        if pos != next_pos:
            # If the current position is not the next position to look for,
            #     then just move on to the next line
            line = fh_in.readline()
            pos += 1
        else:
            # pos == next_pos, hit the header line to look for!
            fh_out.write(line)  # Write the header line

            # The following while loop writes lines until the next header line '>' is seen
            while True:
                line = fh_in.readline()
                pos += 1
                if line.startswith('>') or line == '':
                    # Starts with '>', i.e. a header line, or reach the end of the input file
                    # Break out this while loop for writing the sequence
                    break
                else:
                    # The line is still before the next header line '>'
                    # So it must be part of the DNA sequence to be written into the output file
                    fh_out.write(line)

            # The sequence has been written into the output file
            if len(rand_entry_pos) > 0:
                # If there's more entry position to look for, then update the next_pos
                next_pos = rand_entry_pos.pop(0)
            else:  # len(rand_entry_pos) == 0, no more entry position to look for
                break  # Break out from the main while loop and return

    fh_in.close()
    fh_out.close()


def _sample_fq_pair(
        file1: str,
        file2: str,
        fraction: float,
        output1: str,
        output2: str):
    """
    Randomly subsample the input fastq file pair

    Args:
        file1: path-like
            The input fastq file 1

        file2: path-like
            The input fastq file 2

        fraction:
            Fraction of reads or sequences to be retrieved

        output1: path-like
            The output fastq file 1

        output2: path-like
            The output fastq file 2
    """

    fh_in1 = open(file1, 'r')
    fh_in2 = open(file2, 'r')
    fh_out1 = open(output1, 'w')
    fh_out2 = open(output2, 'w')

    # Count the total number of lines in the fastq file
    i = 0
    while fh_in1.readline() != '':
        i += 1

    # Go back to the beginning of the file, because it needs to go through the file again
    fh_in1.seek(0)

    N = int(i / 4)  # The total count of reads
    n = int(N * fraction)  # The count of reads to be retrieved

    # A random array of int corresponding to which reads to be retrieved
    rand_array = random.sample(list(range(N)), n)
    rand_array.sort()

    # Prepare the iterating variable pos (current line position) and next_pos
    pos = -4
    # 4 lines for each read in fastq, so if the next read is 10th read,
    #     then it starts at 40th line
    next_pos = rand_array.pop(0) * 4
    while len(rand_array) > 0:
        # Read 4 lines at a time
        lines1 = [fh_in1.readline() for _ in range(4)]  # .1.fq
        lines2 = [fh_in2.readline() for _ in range(4)]  # .2.fq
        # Every time read 4 lines, update the current position pos
        pos += 4
        if pos == next_pos:
            # If pos is at the next position, write four lines
            for l in lines1:
                fh_out1.write(l)  # .1.fq
            for l in lines2:
                fh_out2.write(l)  # .2.fq
            # Update the next position
            next_pos = rand_array.pop(0) * 4
        # If pos is not at the next position, then just continue onto the next read

    fh_in1.close()
    fh_in2.close()
    fh_out1.close()
    fh_out2.close()


def _sample_sam(file: str, fraction: float, output: str):
    """
    Randomly subsample the input SAM or BAM file

    Args:
        file: path-like
            The input SAM or BAM file

        fraction:
            Fraction of reads or sequences to be retrieved

        output: path-like
            The output file name
    """
    # The output file format (SAM or BAM) depends on the input format
    if file.endswith('.bam'):
        output_bam = '-b '
    else:  # file ends with '.sam' or others
        output_bam = ''

    call(f"samtools view -s {fraction} {output_bam}{file} > {output}")


def random_sample(
        file: str,
        fraction: float,
        file2: str = '',
        output: str = '',
        output2: str = ''):
    """
    Randomly subsample the reads in the input <file>, and write them into the <output> file

    Supported file formats:
        Fastq (single or paired end)
        Fasta
        SAM/BAM

    The input file could be from other folder,
        but the output file is always written in the current folder

    Supports compressed input files (automatically detects .gz file name)

    Args:
        file: path-like
            The input file (fastq, fasta, sam, bam)

        fraction:
            Fraction of reads or sequences to be randomly sampled

        file2: path-like
            The filename of the second fastq file for paired-end

        output: path-like
            The output file name
            If '', then add the input <file> with prefix 'subset_' as the <output> file name

        output2: path-like
            The output file name for <file2> of paired-end fastq
            If '', then add the input <file2> with prefix 'subset_' as the <output2> file name
    """
    # Unzip the input file if endswith '.gz'
    if file.endswith('.gz'):
        gzip(file, keep=True)
        file = file[:-3]  # Remove '.gz' suffix
        is_gz = True
    else:
        is_gz = False

    if file2.endswith('.gz'):
        gzip(file2, keep=True)
        file2 = file2[:-3]  # Remove '.gz' suffix
        is_gz2 = True
    else:
        is_gz2 = False

    # Determine file type
    if file.endswith('.fa'):
        ftype = 'fa'
    elif (file.endswith('.fq') or file.endswith('.fastq'))\
            and file2 == '':
        ftype = 'fq'
    elif (file.endswith('.fq') or file.endswith('.fastq'))\
            and (file2.endswith('.fq') or file2.endswith('.fastq')):
        ftype = 'fq_paired'
    elif file.endswith('.sam') or file.endswith('.bam'):
        ftype = 'sam'
    else:
        ftype = ''

    if output == '':
        output = 'subset_' + file.split('/')[-1]  # Remove preceding folder path
    if output2 == '' and ftype == 'fq_paired':
        output2 = 'subset_' + file2.split('/')[-1]

    if ftype == 'fa': _sample_fa(file, fraction, output)
    elif ftype == 'fq': _sample_fq(file, fraction, output)
    elif ftype == 'fq_paired': _sample_fq_pair(file, file2, fraction, output, output2)
    elif ftype == 'sam': _sample_sam(file, fraction, output)

    # Remove the temporary unzipped file
    if is_gz: os.remove(file)
    if is_gz2: os.remove(file2)



import os
import itertools
from typing import List, Optional
from .lowlevel import call, check_output


def get_files(
        source: str = '.',
        startswith: str = '',
        endswith: str = '',
        isfullpath: bool = False) -> List[str]:
    """
    Get all files that start with or end with some strings in the source folder

    Args:
        startswith

        endswith

        source: path-like
            The source directory

        isfullpath
            If True, return the full file paths in the list

    Returns:
        A list of file names, or file paths
        If no files found, return an empty list []
    """
    ret = []
    s, e = startswith, endswith
    for path, dirs, files in os.walk(source):
        if path == source:
            ret = [f for f in files if (f.startswith(s) and f.endswith(e))]

    if isfullpath:
        ret = [os.path.join(source, f) for f in ret]

    if ret:
        # Sort the file list so that the order will be
        #     consistent across different OS platforms
        ret.sort()
    return ret


def get_dirs(
        startswith: str = '',
        endswith: str = '',
        source: str = '.',
        isfullpath: bool = False) -> List[str]:
    """
    Similar to 'get_files()' but finds directories
    """

    ret = []
    s, e = startswith, endswith
    for path, dirs, files in os.walk(source):
        if path == source:
            ret = [d for d in dirs if (d.startswith(s) and d.endswith(e))]

    if isfullpath:
        ret = [os.path.join(source, d) for d in ret]

    if ret:
        # Sort the file list so that the order will be
        #     consistent across different OS platforms
        ret.sort()
    return ret


def change_extension(files: List[str], old: str, new: str) -> List[str]:
    """
    Change suffix of a list of file names

    If a file does not have the specified old suffix, then it is not included
    in the returned list

    Args
        files:
            Paths of files

        old:
            The old suffix

        new:
            The new suffix

    Returns:
        A list of file names with new suffices
    """
    ret = []
    for f in files:
        if f.endswith(old):
            ret.append(f[:-len(old)] + new)
    return ret


def change_prefix(files: List[str], old: str, new: str) -> List[str]:
    """
    Change prefix of a list of file names

    If a file does not have the specified old suffix, then it is not included
    in the returned list

    Args
        files:
            Paths of files

        old:
            The old prefix

        new:
            The new prefix

    Returns:
        A list of file names with new suffices
    """
    ret = []
    for f in files:
        if f.startswith(old):
            ret.append(new + f[len(old):])
    return ret


def concat(files: List[str], output: str):
    """
    Args:
        files:
            Each str is a file path

        output: path-like
    """
    call(f"cat {' '.join(files)} > {output}")


def zip_broadcast(*lists):
    """
    This is a broadcasting version of the built-in zip() function
    """
    list_of_lists = []

    # Get the longest length among the input lists
    longest = max(map(len, lists))

    for L in lists:
        if len(L) < longest:
            # If L is not the longest list, use itertools to cycle (i.e. broadcast) the list
            list_of_lists.append(itertools.cycle(L))
        else:  # len(L) == longest
            # If L is the longest list, than just use L to zip
            # len(L) defines the total length of the zipped result
            list_of_lists.append(list(L))

    return list(zip(*list_of_lists))


def tar(dir_: str,
        output: Optional[str] = None,
        verbose: bool = False) -> Optional[str]:
    """
    Args:
        dir_:
            Path to the input dir

        output:
            Path to the output .tar.gz file

        verbose:
            Verbose mode or not

    Returns:
        Path to the output .tar.gz file
    """

    # -c create archive
    # -z gzip
    # -v verbose
    # -f tar.gz filepath

    v = 'v' if verbose else ''

    if dir_.endswith('/'):
        dir_ = dir_[:-1]
    if output is None:
        output = f'{dir_}.tar.gz'
    f = f'-f "{output}"'

    current_dir = os.path.dirname(os.path.abspath(dir_))
    basename = os.path.basename(dir_)

    success = call(f'tar -cz{v} {f} -C "{current_dir}" "{basename}"')

    if success:
        return output
    else:
        return None


def untar(
        file: str,
        archive_path: Optional[str] = None,
        dstdir: Optional[str] = None,
        verbose: bool = False) -> Optional[str]:
    """
    Args:
        file:
            Path to the input .tar.gz file

        archive_path:
            Path to a specific archived file

        dstdir:
            Path to the destination directory in which files are extracted

        verbose:
            Verbose mode or not

    Returns:
        Path to the extracted directory, without the trailing '/'
    """

    # -x extract
    # -z gunzip
    # -v verbose
    # -f tar.gz filepath

    f = f'-f "{file}"'
    v = 'v' if verbose else ''
    a = '' if archive_path is None else f' "{archive_path}"'
    if dstdir is None:
        # dstdir is where the input file locates
        dstdir = os.path.dirname(file)

    success = call(f'tar -zx{v} {f} -C "{dstdir}"{a}')

    if not success:
        return None

    outpath = archive_path if archive_path else tar_list(file=file)[0]

    if outpath.endswith('/'):
        outpath = outpath[:-1]

    return os.path.join(dstdir, outpath)


def tar_list(file: str) -> List[str]:
    """
        Args:
            file:
                Path to the .tar.gz file

        Returns:
            List of archived directories and files
    """
    stdout = check_output(f'tar -tf "{file}"')
    return stdout.strip().split('\n')



from itertools import combinations

IUPAC = {
    ('A', 'C'): 'M',
    ('G', 'T'): 'K',
    ('A', 'G'): 'R',
    ('C', 'T'): 'Y',
    ('C', 'G'): 'S',
    ('A', 'T'): 'W',
    ('C', 'G', 'T'): 'B',
    ('A', 'C', 'G'): 'V',
    ('A', 'G', 'T'): 'D',
    ('A', 'C', 'T'): 'H',
    ('A', 'C', 'G', 'T'): 'N',
}

CODON = {
    'ATT': 'I', 'ATC': 'I', 'ATA': 'I',
    'CTT': 'L', 'CTC': 'L', 'CTA': 'L', 'CTG': 'L', 'TTA': 'L', 'TTG': 'L',
    'GTT': 'V', 'GTC': 'V', 'GTA': 'V', 'GTG': 'V',
    'TTT': 'F', 'TTC': 'F',
    'ATG': 'M',
    'TGT': 'C', 'TGC': 'C',
    'GCT': 'A', 'GCC': 'A', 'GCA': 'A', 'GCG': 'A',
    'GGT': 'G', 'GGC': 'G', 'GGA': 'G', 'GGG': 'G',
    'CCT': 'P', 'CCC': 'P', 'CCA': 'P', 'CCG': 'P',
    'ACT': 'T', 'ACC': 'T', 'ACA': 'T', 'ACG': 'T',
    'TCT': 'S', 'TCC': 'S', 'TCA': 'S', 'TCG': 'S', 'AGT': 'S', 'AGC': 'S',
    'TAT': 'Y', 'TAC': 'Y',
    'TGG': 'W',
    'CAA': 'Q', 'CAG': 'Q',
    'AAT': 'N', 'AAC': 'N',
    'CAT': 'H', 'CAC': 'H',
    'GAA': 'E', 'GAG': 'E',
    'GAT': 'D', 'GAC': 'D',
    'AAA': 'K', 'AAG': 'K',
    'CGT': 'R', 'CGC': 'R', 'CGA': 'R', 'CGG': 'R', 'AGA': 'R', 'AGG': 'R',
    'TAA': '*', 'TAG': '*', 'TGA': '*',
}


def build_aa_dict(codon_dict):
    """
    Invert the key, value relationship in <codon_dict>

    <codon_dict> has codon (str) as key, and single aminoacid (str) as value

    The returned <aa_dict> has aminoacid (str) as key, and list of codons as value

    Args:
        codon_dict: dict
            e.g. {'ATT': 'I', 'ATC': 'I', 'ATA': 'I', ...}

    Returns: aa_dict
        e.g. {'I': ['ATT', 'ATC', 'ATA'], ...}
    """
    aa_dict = {}
    for codon, aa in codon_dict.items():
        aa_dict.setdefault(aa, [])
        aa_dict[aa].append(codon)
    return aa_dict


def generate_wobble_codons(codon_list):
    """
    Given a list of codons, generate all possible wobble codons
    with the third base represented by IUPAC ambiguity codes

    For example,

        ['AAT', 'AAC'] --> ['AAY']

        ['AAT', 'AAC', 'AAA'] --> ['AAY', 'AAM', 'AAW', 'AAH']

        ['AAT', 'AAC', 'GGT', 'GGC'] --> ['AAY', 'GGY']

    Args:
        codon_list: list of str

    Returns: list of str
    """
    wobble_codons = []

    # Build a dict with (the first two bases) as key, and (a list of third bases) as value
    # For example, codon_list = ['AAT', 'AAC',  -->  d = {'AA': ['T', 'C'],
    #                            'GGT', 'GGC']            'GG': ['T', 'C']}
    d = {}
    for codon in codon_list:
        d.setdefault(codon[0:2], [])
        d[codon[0:2]].append(codon[2])

    # Each value is a list, sort the list
    # d = {'AA': ['T', 'C'], --> d = {'AA': ['C', 'T'],
    #      'GG': ['T', 'C']}          'GG': ['C', 'T']}
    d = {key: sorted(val) for key, val in d.items()}

    for first_two, third_bases in d.items():
        if len(third_bases) == 1:
            # No ambiguity at the third base, skip
            continue

        # For example, if third_bases = ['A', 'C', 'G', 'T'],
        # then need to choose k = 2, 3, 4 from the third_bases
        for k in range(2, len(third_bases) + 1):
            for comb in combinations(third_bases, k):
                degenerated_base = IUPAC[comb]
                wobble_codons.append(first_two + degenerated_base)

    return wobble_codons


def build_wobble_codon_dict(codon_dict):
    """
    Build a dictionary with wobble codon (str) as key, and aminoacid (str) as value

    Wobble codon uses IUPAC ambiguity codes for the third base

    Args:
        codon_dict: dict
            e.g. {'ATT': 'I', 'ATC': 'I', 'ATA': 'I', ...}

    Returns: dict
        e.g. {'ATY': 'I', 'ATM': 'I', 'ATW': 'I', 'ATH': 'I', ...}
    """
    ret = {}

    aa_dict = build_aa_dict(codon_dict=codon_dict)

    for aa, codons in aa_dict.items():
        for wobble_codon in generate_wobble_codons(codon_list=codons):
            ret[wobble_codon] = aa

    return ret


def pretty_print_wobble_codon_dict():
    wobble_codon_dict = build_wobble_codon_dict(codon_dict=CODON)

    text = '{\n   '
    cur_prefix, cur_aa = None, None
    for wobble_codon, aa in wobble_codon_dict.items():

        if cur_prefix is None:
            cur_prefix = wobble_codon[:2]
        if cur_aa is None:
            cur_aa = aa

        if wobble_codon[:2] != cur_prefix or aa != cur_aa:
            char0 = '\n    '
            cur_prefix = wobble_codon[:2]
            cur_aa = aa
        else:
            char0 = ' '

        text += f'{char0}{wobble_codon}: {aa},'

    print(text + '\n}')


if __name__ == '__main__':
    pretty_print_wobble_codon_dict()



from typing import List, Optional, Dict, Union
from .lowlevel import printf
from .dataclass import GenericFeature, GffFeature
from .feature_conversion import generic_to_gff_feature


class GffParser:
    """
    File parser for GFF3 format
    """
    def __init__(self, file: str):
        """
        Args:
            file: path-like
        """
        self.__gff = open(file, 'r')
        line1 = self.__gff.readline().rstrip()
        assert line1 == '##gff-version 3', 'The first line is not "##gff-version 3"'

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return

    def __iter__(self):
        self.__gff.seek(0)  # 0 = first character
        self.__gff.readline()
        return self

    def __next__(self):
        r = self.next()
        if r:
            return r
        else:  # r is None
            raise StopIteration

    def next(self) -> Optional[GffFeature]:
        """
        Each line of the GFF3 file has 9 fields

        #   Col	Field        Type
        0   1   seqid        str
        1   2   source       str
        2   3   type         str
        3   4   start        int
        4   5   end          int
        5   6   score        float
        6   7   strand       str ('+', '-')
        7   8   phase        int (0, 1, 2)
        8   9   attributes   str

        Returns: namedtuple of str or int
            9 fields of a feature (i.e. a line)
        """
        line = self.__gff.readline().rstrip()
        if line:
            fields: List[Union[str, int, float]] = line.split('\t')
            for i in (3, 4, 7):
                if fields[i] != '.':
                    fields[i] = int(fields[i])
            if fields[5] != '.':
                fields[5] = float(fields[5])
            return GffFeature._make(fields)
        else:  # line == ''
            return None

    def close(self):
        self.__gff.close()


class GffWriter:
    def __init__(self, file: str, mode: str = 'w'):
        """
        Args:
            file: path-like

            mode:
                The file mode: 'w' or 'a'
        """
        self.__gff = open(file, mode)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return

    def write(self, feature: GffFeature):
        """
        Args:
            feature:
                Containing 9 fields of a line of GFF3 file
        """
        self.__gff.write('\t'.join(map(str, feature)) + '\n')

    def close(self):
        self.__gff.close()


def read_gff(
        file: str,
        as_dict: bool = False) \
        -> Union[List[GffFeature], Dict[str, GffFeature]]:
    """
    Args:
        file: path-like
            The input GFF file

        as_dict:
            If True, returns a dictionary

    Returns: list of GffFeature objects, or dict

        [GffFeature_1, GffFeature_2, ...]

        or

        {
            seqid_1: [GffFeature_1, ...],
            seqid_2: [GffFeature_2, ...], ...
        }
    """
    with GffParser(file) as parser:
        features = [feature for feature in parser]
    if as_dict:
        D = {}
        for f in features:
            D.setdefault(f.seqid, [])
            D[f.seqid].append(f)
        return D
    return features


def write_gff(
        data: Union[List[Union[GffFeature, GenericFeature]],
                    Dict[str, Union[GffFeature, GenericFeature]]],
        file: str):
    """
    Take the data in the format returned by read_gff()
        and write it into a new GFF file

    Accepts GffFeature and GenericFeature objects

    Args:
        data: list of GffFeature objects, or dict

        file: path-like
            The output GFF file
    """
    with GffWriter(file) as writer:
        if type(data) is dict:
            for feature_arr in data.values():
                for feature in feature_arr:
                    if type(feature) is GenericFeature:
                        feature = generic_to_gff_feature(feature)
                    writer.write(feature)
        elif type(data) is list:
            for feature in data:
                if type(feature) is GenericFeature:
                    feature = generic_to_gff_feature(feature)
                writer.write(feature)


def subset_gff(
        file: str,
        seqid: Union[str, List[str]],
        output: str):
    """
    Args:
        file: path-like
            The input GFF file

        seqid:
            Each str is a seqid to be included

        output: path-like
            The output GFF file
    """
    if isinstance(seqid, str):
        seqid = [seqid]

    with GffParser(file) as parser:
        with GffWriter(output) as writer:
            for feature in parser:
                if feature[0] in seqid:
                    writer.write(feature)


def print_gff(
        feature: Optional[GffFeature] = None):
    """
    Pretty print a feature (namedtuple) from GFF or GFF file

    Args:
        feature:
            Containing 9 fields of a feature from GFF or GFF file
    """
    if feature is None:
        text = """\
#   Col	Field       Type
0   1   seqid     string
1   2   source      string
2   3   feature     string
3   4   start       int
4   5   end         int
5   6   score       float
6   7   strand      string ('+', '-')
7   8   frame       int (0, 1, 2)
8   9   attribute   string"""
        printf(text)

    else:
        fields = [
            'seqid    ',
            'source   ',
            'feature  ',
            'start    ',
            'end      ',
            'score    ',
            'strand   ',
            'frame    ',
            'attribute',
        ]
        for i in range(9):
            printf(f"{i}\t{fields[i]}\t{feature[i]}")



from typing import Optional, Tuple


class FastqParser:
    """
    A simple fastq parser that parses each read (four lines) of a fastq file
    """
    def __init__(self, file: str):
        self.__fastq = open(file, 'r')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return

    def __iter__(self):
        self.__fastq.seek(0)
        return self

    def __next__(self):
        r = self.next()
        if r:
             return r
        else:  # r is None:
            raise StopIteration

    def next(self) -> Optional[Tuple[str, str, str, str]]:
        """
        Returns:
            The next read of the fastq file
            If reaches the end of the file, return None
        """
        line1 = self.__fastq.readline().rstrip()
        line2 = self.__fastq.readline().rstrip()
        line3 = self.__fastq.readline().rstrip()
        line4 = self.__fastq.readline().rstrip()
        if line1:
            return line1, line2, line3, line4
        else:  # line1 == ''
            return None

    def close(self):
        self.__fastq.close()


class FastqWriter:
    """
    A simple fastq writer that writes a single read (four lines) each time
    """
    def __init__(self, file: str, mode: str = 'w'):
        """
        Args
            file: path-like

            mode:
                'w' for write or 'a' for append
        """
        self.__fastq = open(file, mode)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return

    def write(self, read: Tuple[str, str, str, str]):
        """
        Args:
            read: tuple of four lines (strings) of each read of fastq
        """
        for line in read:
            self.__fastq.write(line+'\n')

    def close(self):
        self.__fastq.close()


def interleave(fq1: str, fq2: str, fq_out: str):
    """
    Interleave two paired-end fastq files
    This method does not check the header to see
        if the two fastq files are really from paired end reads
    It just interleave 4 lines by 4 lines

    Args:
        fq1: path to fastq1 (.1.fq)

        fq2: path to fastq2 (.2.fq)

        fq_out
    """
    fq1 = open(fq1, 'r')
    fq2 = open(fq2, 'r')
    fq_out = open(fq_out, 'w')

    while True:
        read1 = [fq1.readline().rstrip() for _ in range(4)] # 4 lines -> list
        read2 = [fq2.readline().rstrip() for _ in range(4)] # 4 lines -> list

        # The header line should start with '@'
        if not read1[0].startswith('@') or not read2[0].startswith('@'):
            break

        for line in read1:
            fq_out.write(line+'\n')
        for line in read2:
            fq_out.write(line+'\n')

    fq1.close()
    fq2.close()
    fq_out.close()



from typing import List
from .gtftools import read_gtf, write_gtf
from .dataclass import FeatureArray, GenericFeature


def _pfam_in_orf(
        pfam: GenericFeature, orf: GenericFeature) -> bool:
    """
    Returns:
        Whether the pfam domain is part of the orf
    """
    return pfam.start >= orf.start and \
        pfam.end <= orf.end and \
        pfam.strand == orf.strand and \
        (pfam.start - orf.start) % 3 == 0  # in-frame


def _merge_pfam_arr_to_orf_arr(
        pfam_arr: FeatureArray,
        orf_arr: FeatureArray) -> List[GenericFeature]:
    """
    Returns:
        Each GenericFeature is an ORF with Pfam information appended
    """
    merge_arr = []

    # Note that GenericFeature objects in orf_arr and pfam_arr will be altered
    for orf in orf_arr:
        while len(pfam_arr) > 0:
            pfam = pfam_arr[0]  # the first pfam feature

            # 1) Pfam is contained by the ORF, so add Pfam information into the ORF
            if _pfam_in_orf(pfam, orf):
                pfam_attr = pfam.attributes[:]
                # Add positional information of the pfam domain
                pfam_attr += [('start', pfam.start), ('end', pfam.end), ('strand', pfam.strand)]
                # Convert all pfam attributes [(key, val)] --> str --> 'note' in the ORF
                orf.add_attribute(key='note', val=str(pfam_attr))

                pfam_name = pfam.get_attribute('name')
                orf_name = orf.get_attribute('name')

                orf.set_attribute('name', f'{orf_name} | {pfam_name}')
                pfam_arr.pop(0)

            # 2) The first Pfam is ahead of ORF, go for the next ORF
            elif pfam.start > orf.end:
                break

            # 3) The first Pfam partially overlaps with the ORF,
            #      or is completely before the ORF,
            #      or is on the opposite strand
            else:
                # This is an orphan Pfam not contained by any ORF, add it into the output array
                merge_arr.append(pfam_arr.pop(0))

        merge_arr.append(orf)

    return merge_arr


def merge_pfam_into_orf(pfam: str, orf: str, output: str):
    """
    Args:
        pfam: path-like
            The input GTF containing Pfam annotation

        orf: path-like
            The input GTF containing ORFs

        output: path-like
            The output GTF in which Pfam is merged into ORFs
    """
    pfam_dict = read_gtf(file=pfam, as_dict=True)
    orf_dict = read_gtf(file=orf, as_dict=True)
    merge_dict = {}

    # For each contig, merge Pfam feature array into ORF feature array
    for seqname in orf_dict.keys():

        orf_arr = FeatureArray(
            seqname=seqname,
            genome_size=int(1e6),
            features=orf_dict[seqname],
            circular=False)
        orf_arr.sort()

        pfam_arr = FeatureArray(
            seqname=seqname,
            genome_size=int(1e6),
            features=pfam_dict.get(seqname, []),
            circular=False)
        pfam_arr.sort()

        merge_dict[seqname] = _merge_pfam_arr_to_orf_arr(pfam_arr, orf_arr)

    write_gtf(data=merge_dict, file=output)



import os
from typing import Optional
from .gtftools import GtfWriter
from .fasta import FastaParser, FastaWriter
from .lowlevel import call, _temp, printf


def _remove_extension(file: str) -> str:
    """
    Args:
        file: path-like

    Returns:
        A path without the file extension
    """
    return '.'.join(file.split('.')[:-1])


def _parser_glimmer3_result(file: str, output: str):
    """
    Args:
        file: path-like
            The glimmer3 output prediction file, for example:

            >fasta_header
             orf00001      107      448  +2     3.51
             orf00002      687      800  +3     2.86
             orf00003      725      549  -3     5.63
                          (start)  (end)(frame)(score)

        output: path-like
            The output GTF file
    """
    with open(file) as parser:
        with GtfWriter(output) as writer:
            head = parser.readline().rstrip()[1:]
            for line in parser:
                if line.startswith('>'):
                    head = line.rstrip()[1:]
                else:
                    name, start, end, frame, score = line.rstrip().split()
                    if frame[0] == '-':
                        start, end = end, start
                    writer.write((
                        head,               # seqname
                        '.',                # source
                        'CDS',              # feature
                        start,              # start
                        end,                # end
                        score,              # score
                        frame[0],           # strand
                        frame[1:],          # frame
                        f"name \"{name}\""  # attribute
                    ))


def glimmer3(
        fasta: str,
        output: str,
        linear: bool = True,
        max_overlap: int = 50,
        min_length: int = 110,
        threshold: int = 30,
        entropy_distance: float = 1.15,
        verbose: bool = False,
        log: Optional[str] = None):
    """
    This is a wrapper function of glimmer3, which predicts ORFs

    Glimmer has to run separately on each contig

    For each contig, this function does the following things:
        Find all possible ORFs of the contig    (long-orfs)
        Extract the DNA sequences of long ORFs  (extract)
        Build ICM                               (build-icm)
        Predict ORFs with the ICM               (glimmer3)

    Args:
        fasta: path-like
            The input genome sequences

        output: path-like
            The output GTF file containing predicted ORFs

        linear:
            Linear or circular DNA

        max_overlap:
            Max overlap (bp) between ORFs

        min_length:
            Minimum length of ORFs

        threshold

        entropy_distance:
            Only genes with entropy distance score less than this value will be considered

        verbose:
            If True, print every single command line
            If False, only print the command line for the first sequence of the input <fasta>

        log: path-like
            The log file for stderr
    """
    f = fasta
    for ext in ['.fa', '.fna', '.fasta']:
        if fasta.endswith(ext):
            f = _remove_extension(fasta)

    if log is None:
        log = f"{f}_glimmer3.log"
    log = f" 2>> {log}"

    if os.path.isfile(output):
        os.remove(output)
    output_gtf = open(output, 'a')

    fasta_parser = FastaParser(fasta)

    i = -1
    for head, seq in fasta_parser:
        i += 1
        if verbose or i == 0:
            print_cmd = True
            printf(f"contig: {head}")
        else:
            print_cmd = False

        temp = _temp(prefix='temp', suffix='')
        with FastaWriter(f"{temp}.fa") as writer:
            writer.write(head, seq)

        # --- Find long ORFs --- #
        # cutoff: Only genes with entropy distance score less than <cutoff> will be considered. Default 1.15
        line = ['', '--linear '][linear]
        cmd = f"long-orfs --no_header --cutoff {entropy_distance} {line}{temp}.fa longorfs{log}"
        call(cmd, print_cmd)

        # --- Extract the DNA sequences of long ORFs --- #
        cmd = f"extract {temp}.fa longorfs > train"
        call(cmd, print_cmd)

        # --- Build ICM --- #
        # r: Use the reverse of input strings to build the model
        cmd = f"build-icm -r icm < train"
        call(cmd, print_cmd)

        # --- Predict genes --- #
        # max_olap: Set maximum overlap length to <n>. Overlaps this short or shorter are ignored
        # gene_len: Set minimum gene length to <n>
        # threshold: Set threshold score for calling as gene to n. If the in-frame score >= <n>,
        #   then the region is given a number and considered a potential gene
        cmd = f"glimmer3 --max_olap {max_overlap} --gene_len {min_length} --threshold {threshold} {temp}.fa icm {temp}{log}"
        call(cmd, print_cmd)
        # glimmer3 outputs two files: {temp}.detail and {temp}.predict

        _parser_glimmer3_result(file=f"{temp}.predict", output=f"{temp}.gtf")

        # Append the temp gtf to the output gtf
        with open(f"{temp}.gtf") as fh:
            output_gtf.write(fh.read())

        call(f"rm {temp}.fa longorfs train icm {temp}.detail {temp}.predict {temp}.gtf", print_cmd)

    fasta_parser.close()
    output_gtf.close()



import unittest
from ngslite.dnatools import rev_comp, translate, base_content


class TestDnaTools(unittest.TestCase):

    def test_rev_comp(self):
        self.assertEqual('AGTCagct', rev_comp('agctGACT'))
        self.assertEqual('anAN', rev_comp('NTnt'))

    def test_translate(self):
        dna = 'TAATCGTTTGCAATTCAGGGcTTGATcTACaCTggATTGCCAttcTCTCAAAGTATTaTGCAGGACGGCGTGCgCGTTCCATGtaaaCCTgTCATAACTT'
        prot = '*SFAIQGLIYTGLPFSQSIMQDGVRVPCKPVIT'
        self.assertEqual(prot, translate(dna))

    def test_base_content(self):
        dna = 'AaaCccGgTt'
        for base, content in [
            ('A', 0.3), ('a', 0.3),
            ('C', 0.3), ('c', 0.3),
            ('G', 0.2), ('g', 0.2),
            ('T', 0.2), ('t', 0.2),
        ]:
            self.assertEqual(content, base_content(seq=dna, base=base))


if __name__ == '__main__':
    unittest.main()



from dna_features_viewer import GraphicFeature, GraphicRecord
from .dataclass import Chromosome, GenericFeature
from .genbank_parse import read_genbank


COLOR_DICT = {
    'CDS': '#993366',
    'tRNA': '#228b22',
}


def generic_to_graphic_feature(generic_feature, label_attribute):
    f = generic_feature
    strand = +1 if f.strand == '+' else -1
    label = str(f.get_attribute(label_attribute))
    color = COLOR_DICT.get(f.type, '#ffffff')

    return GraphicFeature(
        start=f.start, end=f.end, strand=strand, label=label, color=color,
        thickness=10, linewidth=0.5, linecolor='#000000', fontdict=None,
        html=None, open_left=True, open_right=False, box_linewidth=1,
        box_color='auto'
    )


def filter_generic_features(generic_features, types, in_types):
    ret = []
    for f in generic_features:
        if (f.type in types) == in_types:
            ret.append(f)
    return ret


def chromosome_to_graphic_record(chromosome, label_attribute, skip_types):

    gen_features = filter_generic_features(
        generic_features=chromosome.features, types=skip_types, in_types=False
    )

    def convert(generic_feature):
        return generic_to_graphic_feature(generic_feature, label_attribute)
    graph_features = list(map(convert, gen_features))

    return GraphicRecord(
        sequence_length=len(chromosome.sequence),
        sequence=chromosome.sequence,
        features=graph_features
    )


def plot_graphic_record(graphic_record, output, fmt, figure_width):
    if figure_width is None:
        figure_width = len(graphic_record.sequence) / 1000
    ax, _ = graphic_record.plot(figure_width=figure_width)
    ax.figure.savefig(output, fmt=fmt, bbox_inches='tight')


def plot_genbank(file, output, fmt='.png', label_attribute='product',
                 hide_types=('gene', 'source'), figure_width=None):

    chromosome = read_genbank(file=file)[0]

    graph_record = chromosome_to_graphic_record(
        chromosome=chromosome, label_attribute=label_attribute,
        skip_types=hide_types
    )

    plot_graphic_record(
        graphic_record=graph_record, output=output, fmt=fmt,
        figure_width=figure_width
    )



import random
import numpy as np
from typing import Union, List


def replace_zero_with(
        np_array: np.ndarray,
        value: Union[int, float]) -> np.ndarray:
    """
    Replaces the 0s in an numpy array with a specified value

    Args:
        np_array:
            1D numpy.ndarray, dtype = np.int or np.float

        value

    Returns: 1D numpy.ndarray, dtype = np.float
    """
    bool_arr = np.equal(np_array, 0)
    val_arr = np.ones((len(np_array), ), dtype=np.float) * value
    return np_array + bool_arr * val_arr


def jitter(
        array: Union[List[Union[int, float]],
                     np.ndarray],
        value: float) -> np.ndarray:
    """
    Add the input <array> with some random numbers
    The random numbers are uniformly distributed between -<value> to +<value>

    Args:
        array:
            An array of numbers, could be list, np.ndarray with int or float data types

        value:
            The max amount of value to be jittered

    Returns: np.ndarray, dtype=np.float32
    """
    # Create a new instance of the input array with np.float32 data type
    array = np.array(array, dtype=np.float32)
    noise = [random.uniform(-value, value) for _ in range(len(array))]
    noise = np.array(noise, dtype=np.float32)
    return array + noise


def densCols(
        x: np.ndarray,
        y: np.ndarray,
        bins: int) -> np.ndarray:
    """
    This is an implementation of the densCols() function in R
    Python just doesn't have this feature!

    Args:
        x:
            x coordinates

        y:
            y coordinates

        bins:
            number of bins

    Returns: 1D numpy array, length = len(x), dtype = np.float64
        Local densities (i.e. counts) of each input 2D point
    """
    # 2D histogram matrix, dtype = np.float64
    hist2D = np.histogram2d(x, y, bins)[0]

    # The bin size is a scalar
    bin_size_x = (max(x) - min(x)) / bins
    bin_size_y = (max(y) - min(y)) / bins

    # For each x and y coordinate, get the position with regard to which bin
    x_pos = (x - min(x)) / bin_size_x
    y_pos = (y - min(y)) / bin_size_y

    # Round it down to integer, so that it could be indexed
    x_pos = x_pos.astype(np.uint, copy=False)
    y_pos = y_pos.astype(np.uint, copy=False)

    # The max of x_pos and y_pos == bins, which could not be indexed (out of bound)
    # Decrease by 1 so that it can be indexed as the last bin
    x_pos[np.argmax(x_pos)] = bins - 1
    y_pos[np.argmax(y_pos)] = bins - 1

    # For each point, find the local density, which is encoded in the hist2D matrix
    densities = hist2D[x_pos, y_pos]

    return densities



from typing import List, Tuple, Optional, Union
from collections import namedtuple
from .dnatools import rev_comp


GtfFeature = namedtuple('GtfFeature', 'seqname source feature start end score strand frame attribute')


GffFeature = namedtuple('GffFeature', 'seqid source type start end score strand phase attributes')


class GenericFeature:
    """
    Instance attributes:
        seqname:
            The unique name of chromosome, genome or contig to which this feature belongs

        type:
            The type of the feature, e.g. 'CDS', 'tRNA', 'misc_feature'

        start:
            1-based inclusive

        end:
            1-based inclusive

        strand:
            {'+' or '-'}

        frame:
            {1, 2, 3}

        attributes:
            [
                ('gene', 'dam'),
                ('locus_tag', 'T4p032')
                ('note', 'produces N6-methyladenine in GATC sites (in modified or non-modified cytosine DNA)'),
                ('codon_start', 1)
            ]

        tags

        regions:
            Default just one region

        partial_start
            Partial 5' end

        partial_end
            Partial 3' end
    """

    def __init__(
            self, seqname: str, type_: str, start: int, end: int, strand: str,
            attributes: Optional[List[Tuple[str, Union[str, int, float]]]] = None,
            tags: Optional[List[str]] = None,
            regions: Optional[List[Tuple[int, int, str]]] = None,
            frame: int = 1, partial_start: bool = False, partial_end: bool = False):

        self.seqname = seqname
        self.type = type_
        self.start = start
        self.end = end
        self.strand = strand
        self.frame = frame
        self.partial_start = partial_start
        self.partial_end = partial_end

        if attributes is None:
            attributes = []
        self.attributes = attributes

        if tags is None:
            tags = []
        self.tags = tags

        if regions is None:
            regions = [(self.start, self.end, self.strand)]
        self.regions = regions

    def __len__(self):
        return self.end - self.start + 1

    def __str__(self):
        ps = ['', ' (partial)'][self.partial_start]
        pe = ['', ' (partial)'][self.partial_end]

        attr_str = ''
        for key, val in self.attributes:
            if type(val) is str:
                attr_str += f"\n    {key}: '{val}'"
            else:  # type(val) is int or float
                attr_str += f"\n    {key}: {val}"

        return f'''\
GenericFeature
  seqname: '{self.seqname}'
  type: '{self.type}'
  start: {self.start}{ps}
  end: {self.end}{pe}
  strand: '{self.strand}'
  frame: {self.frame}
  attributes: {attr_str}
  tags: {','.join(self.tags)}
  regions: {self.regions}'''

    def __repr__(self):
        return f"""GenericFeature(seqname='{self.seqname}', type_='{self.type}'\
, start={self.start}, end={self.end}, strand='{self.strand}', attributes=\
{self.attributes}, tags={self.tags}, regions={self.regions}, frame={self.frame}\
, partial_start={self.partial_start}, partial_end={self.partial_end})"""

    def get_attribute(self, key: str) -> \
            Optional[Union[str, int, float, List[Union[str, int, float]]]]:

        vals = [v for k, v in self.attributes if k == key]
        if len(vals) == 0:
            return None
        elif len(vals) == 1:
            return vals[0]
        else:
            return vals

    def set_attribute(self, key: str, val: Union[str, int, float]):
        """
        If more than one attributes found, then set the first one, and remove the rest
        If no attribute found, then add an attribute with key, val
        """
        new = []  # new attributes list
        done = False  # done setting the attribute or not

        for k, v in self.attributes:
            if k == key:
                if not done:
                    new.append((key, val))
                    done = True
                else:
                    # Already done, don't append the (key, val)
                    #   to the new list, i.e. remove the attribute
                    pass
            else:
                # Keep existing attributes where k != key
                new.append((k, v))

        if not done:
            # No key found, add a new attribute
            new.append((key, val))

        self.attributes = new

    def add_attribute(self, key: str, val: Union[str, int, float]):
        self.attributes.append((key, val))

    def remove_attribute(self, key: str):
        self.attributes = [(k, v) for (k, v) in self.attributes if k != key]


class FeatureArray:
    """
    A list of GenericFeature that behaves like a list, but more than a list
    """

    seqname: str
    genome_size: int
    circular: bool
    __pos: int

    # The main data structure, always remain sorted
    __arr: List[GenericFeature]

    def __init__(
            self, seqname: str, genome_size: int,
            features: Optional[Union[GenericFeature, List[GenericFeature]]] = None,
            circular: bool = False):

        self.seqname = seqname
        self.genome_size = genome_size
        self.circular = circular
        self.__pos = 0

        if features is None:
            features = []
        elif type(features) is GenericFeature:
            features = [features]

        self.__arr = features

        self.sort()

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.__arr[item]
        elif isinstance(item, slice):
            return FeatureArray(seqname=self.seqname, genome_size=self.genome_size, features=self.__arr[item],
                                circular=self.circular)
        else:
            raise TypeError('FeatureArray indices must be integers or slices')

    def __add__(self, other):
        if type(other) is not FeatureArray:
            raise TypeError(f"unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'")

        for attr in ['seqname', 'genome_size', 'circular']:
            assert getattr(self, attr) == getattr(other, attr), f'where attr = \'{attr}\''

        return FeatureArray(
            seqname=self.seqname, genome_size=self.genome_size,
            features=self.__arr + other.__arr, circular=self.circular
        )

    def __iter__(self):
        self.__pos = 0
        return self

    def __next__(self):
        if self.__pos < len(self.__arr):
            r = self.__arr[self.__pos]
            self.__pos += 1
            return r
        else:
            self.__pos = 0
            raise StopIteration

    def __len__(self):
        return len(self.__arr)

    def __str__(self):
        shape = 'circular' if self.circular else 'linear'
        return f'''\
FeatureArray
  seqname: '{self.seqname}'
  genome size: {self.genome_size} bp ({shape})
  number of features: {len(self.__arr)}'''

    def __repr__(self):
        return f"""FeatureArray(seqname={self.seqname}, genome_size=\
{self.genome_size}, features={repr(self.__arr)}, circular={self.circular})"""

    def __wrap(self):
        """
        For circular genomes the feature range can be out of bound (< 1 or > genome_size)
        This method wraps those positions back between 1..genome_size
        """
        size = self.genome_size
        for f in self.__arr:
            if f.start > size:
                f.start -= size
            elif f.start < 1:
                f.start += size

            if f.end > size:
                f.end -= size
            elif f.end < 1:
                f.end += size

            for i in range(len(f.regions)):
                start, end, strand = f.regions[i]

                if start > size:
                    start -= size
                elif start < 1:
                    start += size

                if end > size:
                    end -= size
                elif end < 1:
                    end += size

                f.regions[i] = (start, end, strand)

    def sort(self):
        """
        Sort the features in self.__arr by start positions
        """
        self.__arr = sorted(self.__arr, key=lambda f: f.start)

    def append(self, feature: GenericFeature):
        """
        Append a new GenericFeature
        """
        self.__arr.append(feature)
        self.sort()

    def subset(self, start: int, end: int):
        """
        Subset (or filter) the features in self.__arr without changing their positions

        Args:
            start:
                1-based, inclusive

            end:
                1-based, inclusive
        """

        if not self.circular and end < start:
            raise AssertionError('For linear molecule end position must be greater or equal than start position')

        # Circular range -> Set end to greater than genome_size
        if end < start:
            end += self.genome_size

        arr = []
        for f in self.__arr:
            fs, fe = f.start, f.end
            # Circular feature -> Set feature end to greater than genome_size
            if fe < fs:
                fe += self.genome_size
            # Main criteria to be satisfied
            if start <= fs and fe <= end:
                arr.append(f)
        self.__arr = arr

    def offset(self, offset: int):
        """
        Offset (move) each feature in self.__arr
        """
        size = self.genome_size
        for f in self.__arr:
            f.start += offset
            f.end += offset
            for i, region in enumerate(f.regions):
                start, end, strand = region
                start += offset
                end += offset
                f.regions[i] = (start, end, strand)

            if self.circular:
                self.__wrap()
            # After wrapping every feature will be in-bound

        in_bound = lambda x: 1 <= x.start <= size and 1 <= x.end <= size
        self.__arr = list(filter(in_bound, self.__arr))

    def crop(self, start: int, end: int):
        """
        Crop the features in self.__arr from <start> to <end>,
            which resets positions of features and the genome size

        Args:
            start: int
                1-based, inclusive

            end: int
                1-based, inclusive
        """
        # Covers whole genome -> Do not subset
        if start == 1 and end == self.genome_size:
            return

        if not self.circular and end < start:
            return

        self.subset(start, end)
        self.offset(-start + 1)

        self.genome_size = \
            min(end, self.genome_size) - max(start, 1) + 1

    def reverse(self):
        """
        Reverse the features for reverse complementary of the genome sequence,
            which simply makes start as end, and end as start
        """
        size = self.genome_size
        for f in self.__arr:
            # Simultaneously update start and end
            f.start, f.end = (size - f.end + 1), (size - f.start + 1)
            f.strand = '-' if f.strand == '+' else '+'

            for i in range(len(f.regions)):
                start, end, strand = f.regions[i]
                # Simultaneously update start and end
                start, end = size - end + 1, size - start + 1
                strand = '-' if strand == '+' else '+'
                f.regions[i] = (start, end, strand)
            # Use start position to sort the regions (i.e. exons)
            f.regions = sorted(f.regions, key=lambda x: x[0])

        self.sort()

    def pop(self, index: int = -1):
        """
        Pop out a feature in the self.__arr
        """
        return self.__arr.pop(index)


class Chromosome:
    """
    Instance attributes:
        seqname:
            The unique name of chromosome, genome or contig to which this feature_array belons

        sequence:
            The genome sequence

        feature_array:
            The annotation

        circular:
            Is circular genome or not

        genbank_locus_text:
            The complete LOCUS section of a genbank file
    """

    seqname: str
    sequence: str
    features: FeatureArray
    feature_array: FeatureArray
    circular: bool
    genbank_locus_text: str

    def __init__(
            self, seqname: str, sequence: str, features: FeatureArray,
            circular: bool = False, genbank_locus_text: str = ''):

        self.seqname = seqname
        self.sequence = sequence
        self.features = features
        self.feature_array = self.features
        self.circular = circular
        self.genbank_locus_text = genbank_locus_text

    def crop(self, start: int, end: int):
        """
        Args:
            start: 1-based, inclusive

            end: 1-based, inclusive
        """

        if not self.circular and end < start:
            return

        if start < 1:
            start = 1

        self.sequence = self.sequence[start-1:end]
        self.features.crop(start=start, end=end)

    def reverse(self):
        """
        Reverse the features for reverse complementary of the genome sequence,
            which simply makes start as end, and end as start
        """
        self.sequence = rev_comp(self.sequence)
        self.features.reverse()

    def __repr__(self):
        return f"""Chromosome(seqname='{self.seqname}', sequence='{self.sequence}'\
, features={repr(self.features)}, circular={self.circular}, genbank_locus_text=\
'{self.genbank_locus_text})'"""



def rev_comp(seq: str) -> str:
    """
    Returns reverse complementary sequence of the input DNA string
    """
    comp = {
        'A': 'T',
        'C': 'G',
        'G': 'C',
        'T': 'A',
        'N': 'N',
        'M': 'K',  # M = A C
        'K': 'M',  # K = G T
        'R': 'Y',  # R = A G
        'Y': 'R',  # Y = C T
        'S': 'S',  # S = C G
        'W': 'W',  # W = A T
        'B': 'V',  # B = C G T
        'V': 'B',  # V = A C G
        'D': 'H',  # D = A G T
        'H': 'D',  # H = A C T

        'a': 't',
        'c': 'g',
        'g': 'c',
        't': 'a',
        'n': 'n',
        'm': 'k',
        'k': 'm',
        'r': 'y',
        'y': 'r',
        's': 's',
        'w': 'w',
        'b': 'v',
        'v': 'b',
        'd': 'h',
        'h': 'd'
    }
    return ''.join([comp[base] for base in seq[::-1]])


def translate(dna: str) -> str:
    """
    Translate a DNA sequence
    If the DNA length is not a multiple of 3,
        then leave the last 1 or 2 bases untranslated
    """
    codon = {
        'ATT': 'I', 'ATC': 'I', 'ATA': 'I', 'CTT': 'L',
        'CTC': 'L', 'CTA': 'L', 'CTG': 'L', 'TTA': 'L',
        'TTG': 'L', 'GTT': 'V', 'GTC': 'V', 'GTA': 'V',
        'GTG': 'V', 'TTT': 'F', 'TTC': 'F', 'ATG': 'M',
        'TGT': 'C', 'TGC': 'C', 'GCT': 'A', 'GCC': 'A',
        'GCA': 'A', 'GCG': 'A', 'GGT': 'G', 'GGC': 'G',
        'GGA': 'G', 'GGG': 'G', 'CCT': 'P', 'CCC': 'P',
        'CCA': 'P', 'CCG': 'P', 'ACT': 'T', 'ACC': 'T',
        'ACA': 'T', 'ACG': 'T', 'TCT': 'S', 'TCC': 'S',
        'TCA': 'S', 'TCG': 'S', 'AGT': 'S', 'AGC': 'S',
        'TAT': 'Y', 'TAC': 'Y', 'TGG': 'W', 'CAA': 'Q',
        'CAG': 'Q', 'AAT': 'N', 'AAC': 'N', 'CAT': 'H',
        'CAC': 'H', 'GAA': 'E', 'GAG': 'E', 'GAT': 'D',
        'GAC': 'D', 'AAA': 'K', 'AAG': 'K', 'CGT': 'R',
        'CGC': 'R', 'CGA': 'R', 'CGG': 'R', 'AGA': 'R',
        'AGG': 'R', 'TAA': '*', 'TAG': '*', 'TGA': '*',

        'GTK': 'V'
    }
    dna = dna.upper()
    peptide = []
    for i in range(int(len(dna)/3)):
        aa = codon.get(dna[(i*3):(i*3+3)], 'X')
        peptide.append(aa)
    return ''.join(peptide)


def base_content(seq: str, base: str) -> float:
    """
    Compute the fraction of <base> in the <seq>

    Args
        seq

        base:
            e.g., 'A' gives the A content; 'GC' gives the GC content; 'ACGT' should give 1.0
    """
    seq = seq.upper()
    base = base.upper()

    ret = 0
    for b in set(base):
        ret += seq.count(b) / len(seq)
    return ret



"""
Light-weight functions for next-generation sequencing (NGS) and genomics data analysis

Python API for NGS-related command line tools
Python functions for manipulating NGS and genomics data
"""


__version__ = '1.1.1'


from .cmd_toolkit import bedtools_multicov, bowtie2_mapping, bwa_mapping, \
    metaspades, prokka, trim_galore, orf_finder
from .count import count_reads
from .dataclass import GenericFeature, FeatureArray, Chromosome
from .dnatools import rev_comp, translate, base_content
from .fasta import FastaParser, FastaWriter, read_fasta, write_fasta, subset_fasta
from .fasta_gtf import read_fasta_gtf
from .fastq import FastqParser, FastqWriter, interleave
from .file_conversion import sam_to_bam, bam_to_sam, fq_to_fa, vcf_to_bcf, bcf_to_vcf
from .filetools import get_files, change_extension, change_prefix, concat, zip_broadcast
from .genbank_parse import read_genbank, genbank_to_fasta, genbank_to_gtf
from .genbank_write import write_genbank, make_genbank
from .gfftools import GffParser, GffWriter, read_gff, write_gff, subset_gff, print_gff
from .glimmer import glimmer3
from .gtftools import GtfParser, GtfWriter, read_gtf, write_gtf, subset_gtf, print_gtf
from .hmmer import hmmsearch, hmmbuild, parse_hmmsearch_result, validate_hmm_parse_result
from .locus_extractor import locus_extractor, genbank_locus_extractor
from .lowlevel import call, gzip
from .merge_pfam_orf import merge_pfam_into_orf
from .multiple_sequence_alignment import muscle, simple_phylogeny, draw_tree
from .samtools import SamParser, SamWriter, sort_bam, index_bam, sam_to_indexed_bam, \
    subset_bam_regions, remove_unmapped, decode_flag, encode_flag, filter_sam_by_flag, \
    print_flag, print_sam
from .random import random_sample
from .synteny import synteny
from .vcftools import VcfParser, VcfWriter, bcftools_variant_call, sort_bcf, \
    subset_bcf_regions, print_vcf, unpack_vcf_info



import os
import subprocess
from typing import Optional


def call(cmd: str, print_cmd: bool = True):

    if print_cmd:
        print(f"CMD: {cmd}", flush=True)

    try:
        subprocess.check_call(cmd, shell=True)
    except Exception as inst:
        print(inst, flush=True)


def check_output(cmd: str) -> Optional[str]:
    """
    Wrapper function of subprocess.check_output

    This is meant to act like a function that returns the stdout of a command,
        but not a method that does something, so do not log the cmd

    Returns:
        stdout
        None if call error
    """
    try:
        stdout = subprocess.check_output(cmd, shell=True)
        return str(stdout, 'utf-8')

    except Exception as inst:
        print(inst, flush=True)
        return None


def _temp(prefix: str = 'temp', suffix: str = '') -> str:
    """
    Returns a temp file name that does not exist in the current folder, i.e. temp000.sam

    Args:
        prefix

        suffix:
            Usually the file extension
    """
    i = 0
    while True:
        name = f"{prefix}{i:03}{suffix}"
        if not os.path.exists(name):
            return name
        i += 1


def gzip(file: str, keep: bool = True):
    """
    Call the "gzip" command to zip or unzip files.

    Args:
        file: path-like

        keep:
            Keep the input file or not
    """
    is_gz = file.endswith('.gz')

    decompress = '--decompress ' if is_gz else ''

    if keep:
        stdout = '--stdout '
        if is_gz:
            output = f' > {file[:-3]}'
        else:
            output = f' > {file}.gz'
    else:
        stdout = ''
        output = ''

    call(f"gzip {decompress}{stdout}{file}{output}")


def printf(s: str):
    print(s, flush=True)



from typing import Optional
from .lowlevel import call, gzip, printf
from .dnatools import translate, rev_comp
from .fasta import FastaParser, FastaWriter, read_fasta


def _is_dna(fasta: str) -> bool:
    """
    Returns 'True' if the first sequence of the <fasta> file is a DNA
    """
    with FastaParser(fasta) as parser:
        is_dna = True
        for char in set(parser.next()[1]):
            if not char in ('A', 'C', 'G', 'T'):
                is_dna = False
                break
    return is_dna


def _translate_dna_database(fasta: str) -> str:
    """
    Translate the input fasta file in six frames
    Append ';frame=<frame>' in the header line
    Write a new '<fasta>_translated.fa' file

    Args:
        fasta: path-like

    Returns:
        The written fasta name '<fasta>_translated.fa'
    """
    if fasta.endswith('.fa'):
        output = fasta[:-3] + '_translated.fa'
    elif fasta.endswith('.fasta'):
        output = fasta[:-6] + '_translated.fa'
    else:
        output = fasta + '_translated.fa'

    with FastaParser(fasta) as parser:
        with FastaWriter(output, 'w') as writer:
            for head, seq in parser:
                for frame in (1, 2, 3):
                    new_head = f"{head};frame={frame}"
                    aa_seq = translate(seq[frame-1:])
                    writer.write(new_head, aa_seq)

                for frame in (-1, -2, -3):
                    new_head = f"{head};frame={frame}"
                    rc_seq = rev_comp(seq)
                    aa_seq = translate(rc_seq[-frame-1:])
                    writer.write(new_head, aa_seq)

    return output


def _check_fasta_header(fasta: str) -> bool:
    """
    Fasta headers should NOT contain ' '. This function checks if it's correct
    """
    is_correct = True
    with FastaParser(fasta) as parser:
        for head, seq in parser:
            if ' ' in head:
                is_correct = False
                break
    return is_correct


def hmmsearch(hmm: str, database: str, output: str, cpu: int = 2):
    """
    Args:
        hmm: path-like
            The .hmm file built by the command "hmmbuild",
                e.g. "hmmbuild -o Pfam-A_summary.txt Pfam-A.hmm Pfam-A.seed"
            Also accepts .gz format

        database: path-like
            The fasta database to be searched against
            Also accepts .gz format

        output: path-like
            The output text file reported by the command "hmmsearch"

        cpu:
            Number of CPUs, 2 is the default in hmmsearch
    """
    db_is_gz = False
    if database.endswith('.gz'):
        gzip(database, keep=True)
        database = database[:-len('.gz')]
        db_is_gz = True

    hmm_is_gz = False
    if hmm.endswith('.gz'):
        gzip(hmm, keep=True)
        hmm = hmm[:-len('.gz')]
        hmm_is_gz = True

    db_is_dna = False
    if _is_dna(database):
        database = _translate_dna_database(database)
        db_is_dna = True

    if not _check_fasta_header(database):
        printf(f"Because the database '{database}' contains blank space in headers, hmmsearch is aborted")
        return

    # Run hmmsearch
    call(f"hmmsearch --cpu {cpu} {hmm} {database} > {output}")

    if db_is_gz or db_is_dna:
        call(f"rm {database}")
    if hmm_is_gz:
        call(f"rm {hmm}")


def hmmbuild(seed: str, hmm: str):
    """
    Args:
        seed: path-like
            The input .seed file

        hmm: path-like
            The output .hmm file
    """
    if not hmm.endswith('.hmm'):
        hmm = hmm + '.hmm'
    summary = hmm[:-len('.hmm')] + '_summary.txt'

    call(f"hmmbuild -o {summary} {hmm} {seed}")


def parse_hmmsearch_result(
        file: str, output: str, database: Optional[str] = None):
    """
    Parse the result reported by HMMER, to create a GTF file

    The result text file output by hmmsearch is highly unstructured

    For each hmm query:
        Get <query_name>, <query_accession>, <query_description>

        For each translated contig sequence:
            Get <contig_id>, <contig_length_bp>, <frame>

            For each hit (of the hmm domain) with satisfied threshold:
                1) Get <start_aa> and <end_aa> of amino acid position
                2) Use <start_aa>, <end_aa>, <frame> and <contig_length_bp> to compute <start_bp>, <end_bp>

                Write the domain into GTF as a line, consisting of 9 tab-delimited fields:
                1   seqname     <contig_id>
                2   source      .
                3   feature     CDS
                4   start       <start_bp>
                5   end         <end_bp>
                6   score       .
                7   strand      +/-
                8   frame       0
                9   attribute   name "<query_name>";accession "<query_accession>";description "<query_description>";E_value "<E_value>"

    Args:
        file: path-like
            The text file reported by "hmmsearch"

        output: path-like
            The output GTF file

        database: path-like
            The fasta database used for hmmsearch
            This is used to get the contig length
            If None, then use the contig name in the input <file> to get contig length
    """
    # Read the hmmsearch result
    with open(file, 'r') as fh:
        text = fh.read()

    # Set up a dictionary of contig lengths
    if database:
        with FastaParser(database) as parser:
            contig_length_dict = {head: len(seq) for head, seq in parser}
    else:
        contig_length_dict = {}

    # The output GTF file
    gtf = open(output, 'w')
    gtf_line_count = 0

    # Remove the document header
    text = text[text.find('\n\n') + 2:]

    # Use '//' to split into queries, each query is the result from a particular Pfam-A domain
    all_queries = text.split('\n//\n')[:-1]

    # Remove the sections without hits
    queries = []
    for query in all_queries:
        if not '[No hits detected that satisfy reporting thresholds]' in query:
            queries.append(query)

    # First layer of for loop: Each hmm query is a section
    for query in queries:
        # Get the first three lines
        line1, line2, line3 = query.splitlines()[:3]

        ### Three variables for the query ###
        query_name = line1.split(':')[1].strip()
        query_accession = line2.split(':')[1].strip()
        query_description = line3.split(':')[1].strip()
        ### Three variables for the query ###

        # Get the text between 'Domain annotation for...' and 'Internal pipeline...'
        text = query.split('Domain annotation for each sequence (and alignments):\n')[1].split(
            'Internal pipeline statistics summary:')[0]

        # Use '>> ' to split into translated contigs
        translated_contigs = text.split('>> ')[1:]

        # Second layer of for loop: Each translated contig hit by the query
        for contig in translated_contigs:
            # Very corner case where there's no table
            if '[No individual domains that satisfy reporting thresholds (although complete target did)]' in contig:
                continue

            # Get the table appearing before 'Alignments for each domain'
            table = contig.split('\n\n  Alignments for each domain:')[0]

            # line1 is the header line for the translated contig
            line1 = table.splitlines()[0]
            contig_id, frame = line1.split(';frame=')
            frame = int(frame)
            if contig_length_dict:
                contig_length_bp = contig_length_dict[contig_id]
            else:
                contig_length_bp = int(contig_id.split('len=')[1])

            # From the 4th line, each line is a hit
            hits = table.splitlines()[3:]

            # Third layer of for loop: each hit of a query within the translated contig
            for hit in hits:
                number, satisfied, score, \
                bias, c_Evalue, i_Evalue, \
                hmm_from, hmm_to, symbol_1, \
                ali_from, ali_to, symbol_2, \
                env_from, env_to, symbol_3, acc = hit.split()

                # '!' means: The hit satisfies both per-sequence (contig) and per-domain (query) inclusion thresholds
                if satisfied == '!':
                    # Both <start_aa> and <end_aa> are 1-based index
                    start_aa = int(ali_from)  # aligned from
                    end_aa = int(ali_to)  # aligned to
                    i_Evalue = float(i_Evalue)  # independent E-value

                    # All the bp index are 1-based for GTF files
                    # If frame = 1 or -1, then add (abs(frame) - 1) = 0
                    # Do NOT forget to take absolute value of frame! I spent hours on this bug.
                    start_bp = ((start_aa - 1) * 3 + 1) + (abs(frame) - 1)
                    end_bp = (end_aa * 3) + (abs(frame) - 1)

                    # If + strand, then do nothing to the start_bp and end_bp
                    if frame > 0:
                        strand = '+'
                    # If - strand, then invert it from the - strand positions to the + strand positions
                    else:
                        strand = '-'
                        start_bp, end_bp = contig_length_bp - end_bp + 1, contig_length_bp - start_bp + 1

                    # Get attributes for the particular query, i.e. the Pfam domain
                    attribute = 'name "{}";accession "{}";description "{}";E_value "{}"'.format(
                        query_name, query_accession, query_description, i_Evalue)

                    line = '{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n'.format(
                        contig_id,  # seqname
                        '.',        # source
                        'CDS',      # feature
                        start_bp,   # start
                        end_bp,     # end
                        '.',        # score
                        strand,     # strand
                        0,          # frame
                        attribute)  # attribute

                    gtf.write(line)
                    gtf_line_count += 1

    gtf.close()
    printf('There are totally {} hits exported into the GTF file "{}".'.format(gtf_line_count, output))


def validate_hmm_parse_result(
        gtf: str, dna_database: str, output: str):
    """
    Parsing the HMMER result was very complicated
    In particular, the amino acid positions were converted back to the nucleotide positions
    There could be errors of this conversion
    So I want to verify the nucleotide positions of each predicted domain is correct

    To verify the positions, for each hit in the GTF file:
        Get DNA sequences from the original fasta database
        Translate it into proteins
        See if there's any stop codon *

    After running the script, I saw there are still a few translated sequences that have stop codon
    I went into the HMMER result file and found there was indeed stop codon within the predicted domain

    Args:
        gtf: path-like
            The GTF file exported by parse_hmmsearch_result()

        dna_database: path-like
            The original DNA database (fasta) used from hmmsearch

        output: path-like
            The output text file summarizing the validation result
    """
    contigs = read_fasta(dna_database)

    # Build a dictionary of contigs {header: sequence, ...}
    contigs = {head: seq for head, seq in contigs}

    count_plus = 0
    count_plus_err = 0
    count_minus = 0
    count_minus_err = 0
    with open(gtf, 'r') as gtf_fh:
        for line in gtf_fh:
            seqname, source, feature, start, end, score, strand, frame, attribute = \
                line.rstrip().split('\t')
            start, end = int(start), int(end)

            cds = contigs[seqname][start - 1:end]

            if strand == '+':
                count_plus += 1
            else:
                cds = rev_comp(cds)
                count_minus += 1

            if '*' in translate(cds):
                if strand == '+':
                    count_plus_err += 1
                else:
                    count_minus_err += 1

    with open(output, 'w') as fh:
        text = f"""\
From the GTF file {gtf}:

    There are totally {count_plus} (+) strand CDS in the database {dna_database}, of which:
        {count_plus_err} contains stop codon (*) in the translated sequence.

    There are totally {count_minus} (-) strand CDS in the database {dna_database}, of which:
        {count_minus_err} contains stop codon (*) in the translated sequence."""

        fh.write(text)



from typing import Optional, List, Tuple, Union, Dict
from .lowlevel import call, printf


def bcftools_variant_call(
        ref: str,
        bam: str,
        output: Optional[str] = None,
        max_depth: int = 250,
        haploid: bool = False):
    """
    Wrapper function of "bcftools mpileup" -> BCF (genotype likelihood) -> "bcftools call" -> BCF (variants)

    Args:
        ref: path-like
            The reference fasta file

        bam: path-like
            The alignment BAM file

        output: path-like
            The output BCF file

        max_depth:
            Max per-file sequencing depth; avoids excessive memory usage, default 250

        haploid:
            This is for the option "--ploidy" in "bcftools call"
            If True, set the option "--ploidy 1" for haploid (bacteria)
            If False, ignore the option "--ploidy", the default is diploid
            Command "bcftools call --ploidy ?" prints all available presets: GRCh37, GRCh38, X, Y, 1
    """
    # -Ou: output uncompressed bcf
    # -m: use the default calling method, i.e. alternative model for multiallelic and rare-variant calling
    # -v: output only variant sites
    # -o <file_out>
    # -f <ref_fasta>
    # -d <max_depth>
    # --ploidy <GRCh37|GRCh38|X|Y|1>
    if output is None:
        output = bam[:-4] + '.bcf'
    ploidy = ['', '--ploidy 1 '][haploid]
    call(f"bcftools mpileup -Ou -d {max_depth} -f {ref} {bam} | bcftools call -Ou -m -v {ploidy}-o {output}")


def sort_bcf(
        file: str,
        keep: bool = False):
    """
    Wrapper function of "bcftools sort" to sort BCF and outputs a BCF file

    Args:
        file: path-like

        keep:
            Keep the input file or not
    """
    file_out = f"{file[:-4]}_sorted.{file[-3:]}"
    # -Ou: output uncompressed bcf
    # -o <file_out>
    call(f"bcftools sort -Ou -o {file_out} {file}")
    if not keep:
        call(f"rm {file}")
        call(f"mv {file_out} {file}")


def subset_bcf_regions(
        file: str,
        regions: List[str],
        output: Optional[str] = None,
        keep: bool = True):
    """
    Args:
        file: path-like
            The input BCF file

        regions:
            Each str is a region of the reference genome, e.g.
                chr1            chromosome 1
                chr3:1000-2000  chromosome 3 from 1000th (inclusive) to 2000th (inclusive) base

        output:
            The output BCF file
            If None, add subscript '_subset' to the input <file>

        keep: bool
            If False, delete the input <file> and rename the output as the input <file>
            Overrides the <output> file name
    """
    # Convert regions list into a string
    # ['chr1', 'chr2:1001-2000'] -> ' chr1,chr2:1001-2000'
    regions = ' ' + ','.join(regions)  # comma-separated regions

    if output is None:
        output = file[:-4] + '_subset.bcf'

    # -Ou: output uncompressed bcf
    # -o <file_out>
    call(f"bcftools view -Ou -o {output} {file}{regions}")

    if not keep:
        call(f"rm {file}")
        call(f"mv {output} {file}")


class VcfParser:
    """
    A VCF file parser for VCFv4.3
    """
    def __init__(self, file: str):
        """
        The header section is parsed upon instantiation

        Args:
            file: path-like
        """
        self.__vcf = open(file, 'r')
        header = ''
        while True:
            # Get the current position
            pos = self.__vcf.tell()
            # Readline and move on to the next position
            line = self.__vcf.readline()
            if line.startswith('#'):
                header = header + line
            else:
                # If reaching the alignment section, that is,
                #   the header section has been parsed completely,
                #   then go back one line and break out the loop
                self.__vcf.seek(pos)
                # Remember where the first line of data is
                self.pos_0 = pos
                break
        # Store the header string
        self.header = header

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return

    def __iter__(self):
        self.__vcf.seek(self.pos_0)
        return self

    def __next__(self):
        r = self.next()
        if r:
            return r
        else:  # r is None
            self.__vcf.close()
            raise StopIteration

    def next(self) -> Optional[Tuple[Union[str, int]]]:
        """
        Each line of VCF file has 8 fixed, mandatory fields

        #   Col	Field   Type    Description
        0   1   CHROM   String  The name of the sequence (typically a chromosome)
        1   2   POS     Int     The 1-based position of the variation on the given sequence
        2   3   ID      String  The identifier of the variation
        3   4   REF     String  The reference base (or bases in the case of an indel) at the given position on the given reference sequence
        4   5   ALT     String  The list (separated by ,) of alternative alleles at this position
        5   6   QUAL    Int     A quality score associated with the inference of the given alleles.
        6   7   FILTER  String  A flag indicating which of a given set of filters the variation has passed
        7   8   INFO    String  An extensible list of key-value pairs (fields) describing the variation, for example, NS=2;DP=10;AF=0.333,0.667;AA=T;DB

        Returns: tuple of str or int
            8 fields of a variant (i.e. a line) plus optional fields
        """
        line = self.__vcf.readline().rstrip()
        if line:
            fields: List[Union[str, int]] = line.split('\t')
            for i in [1, 5]:
                fields[i] = int(fields[i])
            return tuple(fields)
        else:  # line == ''
            return None

    def close(self):
        self.__vcf.close()


class VcfWriter:
    def __init__(self, file: str, header: str, mode: str = 'w'):
        """
        The header section should be written in upon instantiation

        Args:
            file: path-like

            header:
                The header section

            mode:
                File mode: 'w' or 'a'
        """
        self.__vcf = open(file, mode)
        if not header == '':
            self.__vcf.write(header.rstrip() + '\n')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return

    def write(self, variant: Tuple[Union[str, int]]):
        """
        Args:
            variant: tuple
                Containing at least 8 fields of a line of VCF file
        """
        self.__vcf.write('\t'.join(map(str, variant)) + '\n')

    def close(self):
        self.__vcf.close()


def print_vcf(var: Tuple[Union[str, int]] = None):
    """
    Pretty print a variant (tuple) from vcf file

    Args:
        var: tuple or list of str or int
            Containing (at least) 8 fields of a variant from vcf file
    """
    if var is None:
        text = ''''\
#   Col	Field   Type    Description
0   1   CHROM   String  The name of the sequence (typically a chromosome)
1   2   POS     Int     The 1-based position of the variation on the given sequence
2   3   ID      String  The identifier of the variation
3   4   REF     String  The reference base (or bases in the case of an indel) at the given position on the given reference sequence
4   5   ALT     String  The list (separated by ,) of alternative alleles at this position
5   6   QUAL    Int     A quality score associated with the inference of the given alleles.
6   7   FILTER  String  A flag indicating which of a given set of filters the variation has passed
7   8   INFO    String  An extensible list of key-value pairs (fields) describing the variation, for example, NS=2;DP=10;AF=0.333,0.667;AA=T;DB
8   9                   Optional fields...'''
        printf(text)

    elif type(var) is tuple or type(var) is list:
        fields = ['CHROM ', 'POS   ', 'ID    ', 'REF   ', 'ALT   ',
                  'QUAL  ', 'FILTER', 'INFO  ']
        for i in range(8):
            printf(f"{i}\t{fields[i]}\t{var[i]}")
        if len(var) > 8:
            for i in range(8, len(var)):
                printf(f"{i}\t     \t{var[i]}")


def unpack_vcf_info(var: Tuple[Union[str, int]]) -> Dict[str, str]:
    """
    Args:
        var: tuple or list of str or int
            Containing (at least) 8 fields of a variant from vcf file

    Return: dict
        Key, value pairs from the INFO (column 7) of a variant
        Values still str, not converted to int or float
    """
    def func(x: str) -> List[str]:
        return x.split('=')
    return {key: val for key, val in map(func, var[7].split(';'))}



from .dataclass import GenericFeature, GtfFeature, GffFeature


def gtf_to_generic_feature(gtf_feature: GtfFeature) -> GenericFeature:

    assert type(gtf_feature) is GtfFeature

    f = gtf_feature

    attr_list = []
    for a in f.attribute.split(';'):
        key = a.split(' "')[0]
        val = a[len(key)+2:-1]
        attr_list.append((key, val))

    return GenericFeature(
        seqname=f.seqname,
        type_=f.feature,
        start=f.start,
        end=f.end,
        strand=f.strand,
        attributes=attr_list,
        frame=f.frame + 1
    )


def gff_to_generic_feature(gff_feature: GffFeature) -> GenericFeature:

    assert type(gff_feature) is GffFeature

    f = gff_feature
    items = [item for item in f.attributes.split(';') if item]

    attr_list = []
    for item in items:
        key, val = item.split('=')
        attr_list.append((key, val))

    return GenericFeature(
        seqname=f.seqid,
        type_=f.type,
        start=f.start,
        end=f.end,
        strand=f.strand,
        attributes=attr_list,
        frame=1 if f.phase == '.' else f.phase + 1
    )


def generic_to_gtf_feature(generic_feature: GenericFeature) -> GtfFeature:
    """
    Covert GenericFeature to GtfFeature (namedtuple)

    Args:
        generic_feature: GenericFeature object
    """
    assert type(generic_feature) is GenericFeature

    f = generic_feature

    # Pack attributes into a single line of str
    attr_str = ''
    for key, val in f.attributes:
        if type(val) is int or type(val) is float:
            attr_str += f"{key} {val};"
        else:  # type(val) is str -> Add quote ""
            val = val.replace(';', '<semicolon>')  # semicolon is not allowed in the GTF attribute field
            attr_str += f"{key} \"{val}\";"

    return GtfFeature(
        seqname=f.seqname,
        source='.',
        feature=f.type,
        start=f.start,
        end=f.end,
        score='.',
        strand=f.strand,
        frame=f.frame - 1,  # GTF frame is 0, 1, 2
        attribute=attr_str[:-1]  # Remove trailing ';'
    )


def generic_to_gff_feature(generic_feature: GenericFeature) -> GffFeature:
    """
    Covert GenericFeature to GffFeature (namedtuple)

    Args:
        generic_feature: GenericFeature object
    """
    assert type(generic_feature) is GenericFeature

    f = generic_feature

    # Pack attributes into a single line of str
    attr_str = ''
    for key, val in f.attributes:
        # Escape characters not allowed
        if type(val) is str:
            val = val.replace(';', '%3B').replace('=', '%3D')
        attr_str += f"{key}={val};"

    return GffFeature(
        seqid=f.seqname,
        source='.',
        type=f.type,
        start=f.start,
        end=f.end,
        score='.',
        strand=f.strand,
        phase=f.frame - 1,  # GFF frame is 0, 1, 2
        attributes=attr_str[:-1]  # Remove trailing ';'
    )



import shutil
import unittest
from ngslite.fasta import FastaParser
from ngslite.dataclass import GenericFeature
from ngslite.genbank_parse import get_seqname, get_circular, \
    split_features_text, get_feature_type, get_feature_location, \
    get_feature_attributes, get_sequence, is_valid_first_line_of_feature, \
    construct_chromosome, genbank_to_fasta, genbank_to_gff, genbank_to_gtf, \
    read_genbank, GenbankTextParser
from .setup_dirs import setup_dirs


locus_text = '''\
LOCUS       NC_000866      168903 bp    DNA     linear   PHG 15-FEB-2017
DEFINITION  Enterobacteria phage T4, complete genome.
ACCESSION   NC_000866
KEYWORDS    .
SOURCE      Enterobacteria phage T4 (T4)
  ORGANISM  Enterobacteria phage T4
            Viruses; dsDNA viruses, no RNA stage; Caudovirales; Myoviridae;
            T4-like viruses.
REFERENCE   1  (bases 1 to 168903)
  AUTHORS   Miller,E.S., Kutter,E., Mosig,G., Arisaka,F., Kunisawa,T. and
            Ruger,W.
  TITLE     Bacteriophage T4 Genome
  JOURNAL   Microbiol. Mol. Biol. Rev. 67 (1), 86-156 (2003)
  PUBMED    12626685
REFERENCE   2  (bases 80196 to 80618)
  AUTHORS   Shcherbakov,V., Granovsky,I., Plugina,L., Shcherbakova,T.,
            Sizova,S., Pyatkov,K., Shlyapnikov,M. and Shubina,O.
  TITLE     Focused genetic recombination of bacteriophage t4 initiated by
            double-strand breaks
  JOURNAL   Genetics 162 (2), 543-556 (2002) MEDLINE 22286362
  PUBMED    12399370
'''


features_text = '''\
FEATURES             Location/Qualifiers
     source          1..168903
                     /organism="Enterobacteria phage T4"
                     /mol_type="genomic DNA"
                     /db_xref="taxon:10665"
     mRNA            join(484000..485000,486000..487000,488000..489000)
                     /gene=60
                     /locus_tag="T4p003"
                     /note="topoisomerase II in T4 and related phages an
                     insertion splits the large Topo II subunit in two parts
                     (gp60 and gp39); a 50-bp segment is inserted in T4 gene 60
                     that forms an mRNA secondary structure that is
                     translationally bypassed"
                     /product="gp60 topoisomerase II, large subunit,
                     C-terminalregion"
                     /protein_id="NP_049618.1"
                     /db_xref="GI:9632675"
                     /db_xref="GeneID:1258779"
     CDS             complement(join(<484000..485000,486000..487000,
                     488000..>489000))
                     /gene=60
                     /locus_tag="T4p003"
                     /note="topoisomerase II in T4 and related phages an
                     insertion splits the large Topo II subunit in two parts
                     (gp60 and gp39); a 50-bp segment is inserted in T4 gene 60
                     that forms an mRNA secondary structure that is
                     translationally bypassed"
                     /codon_start=1
                     /transl_table=11
                     /product="gp60 topoisomerase II, large subunit,
                     C-terminalregion"
                     /protein_id="NP_049618.1"
                     /db_xref="GI:9632675"
                     /db_xref="GeneID:1258779"
                     /translation="MKFVKIDSSSVDMKKYKLQNNVRRSIKSSSMNYANVAIMTDADH
                     DGLGSIYPSLLGFFSNWPELFEQGRIRFVKTPVIIAQVGKKQEWFYTVAEYESAKDAL
                     PKHSIRYIKGLGSLEKSEYREMIQNPVYDVVKLPENWKELFEMLMGDNADLRKEWMSQ
                     "
'''


feature_text_1 = '''\
     mRNA            join(484000..485000,486000..487000,488000..489000)
                     /gene=60
                     /locus_tag="T4p003"
                     /note="topoisomerase II in T4 and related phages an
                     insertion splits the large Topo II subunit in two parts
                     (gp60 and gp39); a 50-bp segment is inserted in T4 gene 60
                     that forms an mRNA secondary structure that is
                     translationally bypassed"
                     /product="gp60 topoisomerase II, large subunit,
                     C-terminalregion"
                     /protein_id="NP_049618.1"
                     /db_xref="GI:9632675"
                     /db_xref="GeneID:1258779"
'''


feature_text_2 = '''\
     CDS             complement(join(<484000..485000,486000..487000,
                     488000..>489000))
                     /gene=60
                     /locus_tag="T4p003"
                     /note="topoisomerase II in T4 and related phages an
                     insertion splits the large Topo II subunit in two parts
                     (gp60 and gp39); a 50-bp segment is inserted in T4 gene 60
                     that forms an mRNA secondary structure that is
                     translationally bypassed"
                     /codon_start=1
                     /transl_table=11
                     /product="gp60 topoisomerase II, large subunit,
                     C-terminalregion"
                     /protein_id="NP_049618.1"
                     /db_xref="GI:9632675"
                     /db_xref="GeneID:1258779"
                     /translation="MKFVKIDSSSVDMKKYKLQNNVRRSIKSSSMNYANVAIMTDADH
                     DGLGSIYPSLLGFFSNWPELFEQGRIRFVKTPVIIAQVGKKQEWFYTVAEYESAKDAL
                     PKHSIRYIKGLGSLEKSEYREMIQNPVYDVVKLPENWKELFEMLMGDNADLRKEWMSQ
                     "
'''


origin_text = '''\
ORIGIN      
        1 tgaatggttc attatctaaa aacttgttag caaccttaga tcaggttgaa gctgaggata
       61 tcgcacataa ccaactagaa ttagaacggg ctggtctgga tgatcgagcg tttggtggtg
      121 tta
'''


class TestGenbankParse(unittest.TestCase):

    def setUp(self):
        self.testdir, self.datadir, self.workdir, self.outdir = setup_dirs(__file__)

    def tearDown(self):
        shutil.rmtree(self.workdir)
        shutil.rmtree(self.outdir)

    def test_get_seqname(self):
        self.assertEqual('NC_000866', get_seqname(locus_text))

    def test_get_circular(self):
        self.assertFalse(get_circular(locus_text))

    def test_split_features_text(self):
        list_ = split_features_text(features_text)
        self.assertEqual(feature_text_1, list_[1])
        self.assertEqual(feature_text_2, list_[2])

    def test_get_feature_type(self):
        self.assertEqual('mRNA', get_feature_type(feature_text_1))
        self.assertEqual('CDS', get_feature_type(feature_text_2))

    def test_get_feature_location(self):
        start, end, strand, regions, partial_start, partial_end = get_feature_location(feature_text_1)
        self.assertEqual(484000, start)
        self.assertEqual(489000, end)
        self.assertEqual('+', strand)
        self.assertEqual((484000, 485000, '+'), regions[0])
        self.assertEqual((486000, 487000, '+'), regions[1])
        self.assertEqual((488000, 489000, '+'), regions[2])
        self.assertFalse(partial_start)
        self.assertFalse(partial_end)

        start, end, strand, regions, partial_start, partial_end = get_feature_location(feature_text_2)
        self.assertEqual(484000, start)
        self.assertEqual(489000, end)
        self.assertEqual('-', strand)
        self.assertEqual((484000, 485000, '-'), regions[0])
        self.assertEqual((486000, 487000, '-'), regions[1])
        self.assertEqual((488000, 489000, '-'), regions[2])
        self.assertTrue(partial_start)
        self.assertTrue(partial_end)

    def test_get_feature_attributes(self):
        attr = get_feature_attributes(feature_text_2)
        self.assertEqual(attr[0], ('gene', 60))
        self.assertEqual(attr[1], ('locus_tag', 'T4p003'))
        self.assertEqual(attr[2], ('note', 'topoisomerase II in T4 and related phages an \
insertion splits the large Topo II subunit in two parts \
(gp60 and gp39); a 50-bp segment is inserted in T4 gene 60 \
that forms an mRNA secondary structure that is \
translationally bypassed'))
        self.assertEqual(attr[3], ('codon_start', 1))
        self.assertEqual(attr[4], ('transl_table', 11))
        self.assertEqual(attr[5], ('product', 'gp60 topoisomerase II, large subunit, C-terminalregion'))
        self.assertEqual(attr[6], ('protein_id', 'NP_049618.1'))
        self.assertEqual(attr[7], ('db_xref', 'GI:9632675'))
        self.assertEqual(attr[8], ('db_xref', 'GeneID:1258779'))
        self.assertEqual(attr[9], ('translation', 'MKFVKIDSSSVDMKKYKLQNNVRRSIKSSSMNYANVAIM\
TDADHDGLGSIYPSLLGFFSNWPELFEQGRIRFVKTPVIIAQVGKKQEWFYTVAEYESAKDALPKHSIRYIKGLGSLEKSEYREMIQNPV\
YDVVKLPENWKELFEMLMGDNADLRKEWMSQ'))

    def test_get_sequence(self):
        self.assertEqual(
            'tgaatggttcattatctaaaaacttgttagcaaccttagatcaggttgaagctgagga\
tatcgcacataaccaactagaattagaacgggctggtctggatgatcgagcgtttggtggtgtta',
            get_sequence(origin_text)
        )

    def test_is_valid_first_line_of_feature(self):
        valid_lines = [
            '     mRNA            join(1..10,100..110)'
            '     CDS             complement(1..10)'
            '     CDS             <1..10'
        ]

        invalid_lines = [
            '                     ',
            '     mRNA CDS        1..10'
        ]

        for line in valid_lines:
            self.assertTrue(is_valid_first_line_of_feature(line))

        for line in invalid_lines:
            self.assertFalse(is_valid_first_line_of_feature(line))

    def test_construct_chromosome(self):
        expected_feature = GenericFeature(
            seqname='NC_000866',
            type_='CDS',
            start=2458,
            end=2990,
            strand='-',
            attributes=[
                ('gene', '60'),
                ('locus_tag', 'T4p003'),
                ('note', 'topoisomerase II in T4 and related phages an insertion \
splits the large Topo II subunit in two parts (gp60 and gp39); a 50-bp segment is \
inserted in T4 gene 60 that forms an mRNA secondary structure that is translationally bypassed'),
                ('codon_start', 1),
                ('transl_table', 11),
                ('product', 'gp60 topoisomerase II, large subunit, C-terminal region'),
                ('protein_id', 'NP_049618.1'),
                ('db_xref', 'GeneID:1258779'),
                ('translation', 'MKFVKIDSSSVDMKKYKLQNNVRRSIKSSSMNYANVAIMTDADHD\
GLGSIYPSLLGFFSNWPELFEQGRIRFVKTPVIIAQVGKKQEWFYTVAEYESAKDALPKHSIRYIKGLGSLEKSEYRE\
MIQNPVYDVVKLPENWKELFEMLMGDNADLRKEWMSQ')
            ],
            regions=[(2458, 2802, '-'), (2853, 2990, '-')],
            frame=1,
            partial_start=False,
            partial_end=False
        )

        with GenbankTextParser(file=f'{self.datadir}/NC_000866.gbk') as parser:
            genbank_text = parser.next()
            chromosome = construct_chromosome(genbank_text=genbank_text)

        for feature in chromosome.features:
            locus_tag = feature.get_attribute('locus_tag')
            if feature.type == 'CDS' and locus_tag == 'T4p003':
                self.assertEqual(repr(expected_feature), repr(feature))
                break

    def test_genbank_to_fasta(self):
        fname = 'NC_000866'
        genbank_to_fasta(
            file=f'{self.datadir}/{fname}.gbk',
            output=f'{self.outdir}/{fname}.fna'
        )

        parser1 = FastaParser(f'{self.outdir}/{fname}.fna')
        parser2 = FastaParser(f'{self.datadir}/{fname}.fna')

        for item1, item2 in zip(parser1, parser2):
            _, seq1 = item1
            _, seq2 = item2
            self.assertEqual(seq1.lower(), seq2.lower())

        parser1.close()
        parser2.close()

    def test_genbank_to_gtf(self):
        fname = 'NC_000866'
        genbank_to_gtf(
            file=f'{self.datadir}/{fname}.gbk',
            output=f'{self.outdir}/{fname}.gtf',
            skip_types=''
        )

    def test_genbank_to_gff(self):
        fname = 'NC_000866'
        genbank_to_gff(
            file=f'{self.datadir}/{fname}.gbk',
            output=f'{self.outdir}/{fname}.gff3',
            skip_types=''
        )

    def test_read_genbank(self):
        target_feature = GenericFeature(
            seqname='NC_000866',
            type_='CDS',
            start=2458,
            end=2990,
            strand='-',
            attributes=[
                ('gene', '60'),
                ('locus_tag', 'T4p003'),
                ('note', 'topoisomerase II in T4 and related phages an insertion \
splits the large Topo II subunit in two parts (gp60 and gp39); a 50-bp segment is \
inserted in T4 gene 60 that forms an mRNA secondary structure that is translationally bypassed'),
                ('codon_start', 1),
                ('transl_table', 11),
                ('product', 'gp60 topoisomerase II, large subunit, C-terminal region'),
                ('protein_id', 'NP_049618.1'),
                ('db_xref', 'GeneID:1258779'),
                ('translation', 'MKFVKIDSSSVDMKKYKLQNNVRRSIKSSSMNYANVAIMTDADHD\
GLGSIYPSLLGFFSNWPELFEQGRIRFVKTPVIIAQVGKKQEWFYTVAEYESAKDALPKHSIRYIKGLGSLEKSEYRE\
MIQNPVYDVVKLPENWKELFEMLMGDNADLRKEWMSQ')
            ],
            regions=[(2458, 2802, '-'), (2853, 2990, '-')],
            frame=1,
            partial_start=False,
            partial_end=False
        )

        fname = 'NC_000866'
        chromosomes = read_genbank(file=f'{self.datadir}/{fname}.gbk')
        for chromosome in chromosomes:
            for feature in chromosome.features:
                locus_tag = feature.get_attribute('locus_tag')
                if feature.type == 'CDS' and locus_tag == 'T4p003':
                    self.assertEqual(repr(feature), repr(target_feature))
                    break


if __name__ == '__main__':
    unittest.main()



from datetime import date
from .fasta_gtf import read_fasta_gtf
from .dnatools import translate, rev_comp
from .genbank_parse import GenbankText


class GenbankTextWriter:
    def __init__(self, file: str, mode: str = 'w'):
        """
        Args
            file: path-like

            mode:
                'w' for write or 'a' for append
        """
        self.__gbk = open(file, mode)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def write(self, genbank_text: GenbankText):
        """
        Args:
            genbank_text: namedtuple GenbankItem
        """
        for i in range(3):
            self.__gbk.write(genbank_text[i].rstrip() + '\n')
        self.__gbk.write('//\n')

    def close(self):
        self.__gbk.close()


def get_today() -> str:
    """
    Returns: str,
        Today, e.g. '20-JUN-2006'
    """
    today = date.today()
    day = today.strftime('%d')
    month = today.strftime('%b').upper()
    year = today.strftime('%Y')
    return f"{day}-{month}-{year}"


def wrap_line_by_word(line, length, indent, sep=' ', keep_sep=False) -> str:
    """
    Wraps a line (not containing any '\n') by word
    The returned line does not end with '\n'

    Args:
        line: str

        length: int

        indent: int

        sep: str
            The separator of words, usually ' '

        keep_sep: bool
            When wrapping, whether or not to keep the separator attached to the word

    Returns: str
    """
    words = line.split(sep)
    if keep_sep:
        # separator always attached to the word (except the last word)
        for i, word in enumerate(words[:-1]):
            words[i] = word + sep

    indent = ' ' * indent
    text = indent + words[0]

    for word in words[1:]:
        last_line = text.split('\n')[-1]

        if keep_sep:
            # Separator is always attached to the word, no need to consider the separator's length
            if len(last_line) + len(word) <= length:
                text += word
            else:
                text += '\n' + indent + word

        else:
            # Need to consider the separator's length
            if len(last_line) + 1 + len(word) <= length:
                text += sep + word
            else:
                text += '\n' + indent + word

    return text


def wrap_line_by_char(line, length, indent) -> str:
    """
    Wraps a line (not containing any '\n') by character
    The returned line does not end with '\n'

    Args:
        line: str

        length: int

        indent: int

    Returns: str
    """
    text = ' ' * indent
    for char in line:
        last_line = text.split('\n')[-1]
        if len(last_line) + 1 <= length:
            text += char
        else:
            text += '\n' + ' '*indent + char
    return text


def generic_feature_to_genbank_text(feature):
    """
    Use the information in a GenericFeature object to write a genbank feature
        text section, for example:

    `     CDS             complement(join(<484868..485129,485184..485520,
    `                     485576..>486058))
    `                     /locus_tag="GN000052.00_006979"
    `                     /codon_start=3
    `                     /product="hypothetical protein"
    `                     /protein_id="ncbi:GN000052.00_006979-T1"
    `                     /translation="PAPQAPAPQAPAPQAPAPQAPAPHAPAPHAPAPHAPAQHAPAQQ
    `                     ASLQGAPAHQVQMPLGVPPPTPSSLRHIQSAIIREEPVYPEPVHRTQSPSPRTHGTKK
    `                     RANKEYTADSESGPSWGFDDVYGDGFKLADEDKTDLVELTEAQAKNVKTLGLAAGGKL
    `                     IQDIYKDPFPAAIWNHAGARILHVHMIDPESCERVTHIVPQPPPMAVEEYVKSGGQFF
    `                     VVEEKVDERLDGGDFDNVKSVSQMDQHLGITTEPEFDPKKPKMCTTCERRLCDCIIRP
    `                     CNHQFCNVCIKRLDEAGDTEQSVQQARHWKCPTCNSPVSHVAGFSAPMNLPGEERLLR
    `                     TKVPVHVLKIEDGRMRLSSMQTSRV"

    Args:
        feature: GenericFeature object

    Returns: str
    """
    f = feature

    lines = []

    # First line: regions
    regions = [f'{start}..{end}' for start, end, _ in f.regions]
    regions = ','.join(regions)
    if f.partial_start:
        regions = '<' + regions
    if f.partial_end:
        # Insert '>' after the last '..'
        pos = regions.rfind('..') + 2
        regions = regions[:pos] + '>' + regions[pos:]
    if len(f.regions) > 1:
        regions = f'join({regions})'
    if f.strand == '-':
        regions = f'complement({regions})'

    lines.append(regions)

    # Second to the last line
    for key, val in f.attributes:
        if type(val) is int or type(val) is float:
            lines.append(f'/{key}={val}')
        else:
            lines.append(f'/{key}="{val}"')

    text = ''
    for i, line in enumerate(lines):
        if i == 0:
            # Wrap the first line (regions) by ',' and keep ','
            text += wrap_line_by_word(line, length=79, indent=21, sep=',', keep_sep=True)
        elif line.startswith('/translation'):
            text += wrap_line_by_char(line, length=79, indent=21)
        else:
            text += wrap_line_by_word(line, length=79, indent=21, sep=' ', keep_sep=False)
        text += '\n'

    # Fill in the feature type, e.g. 'CDS'
    head = ' '*5 + f.type
    text = head + text[len(head):]

    return text


def make_header(
        molecule,
        length,
        shape,
        ACCESSION,
        DEFINITION,
        KEYWORDS,
        SOURCE,
        ORGANISM,
        division='ENV') -> str:
    """
    Create a header section of a genbank file

    Args:
        molecule: str, e.g. 'DNA', 'mRNA'
        length: int
        shape: str, 'linear' or 'circular'
        ACCESSION: str
        DEFINITION: str
        KEYWORDS: str
        SOURCE: str
        ORGANISM: str
        division: str,
            The three-letter tag for one of the 18 divisions in the GenBank database.
            Default 'ENV' for environmental sampling sequences

    Returns: str
        Text of genbank header
    """

    # Replace all '\n' with ' ' to become single line
    ACCESSION  = ACCESSION.replace('\n', ' ')
    DEFINITION = DEFINITION.replace('\n', ' ')
    KEYWORDS   = KEYWORDS.replace('\n', ' ')
    SOURCE     = SOURCE.replace('\n', ' ')

    # Wrap and then remove the indent of the first line
    ACCESSION  = wrap_line_by_word(ACCESSION, length=79, indent=12).lstrip()
    DEFINITION = wrap_line_by_word(DEFINITION, length=79, indent=12).lstrip()
    KEYWORDS   = wrap_line_by_word(KEYWORDS, length=79, indent=12).lstrip()
    SOURCE     = wrap_line_by_word(SOURCE, length=79, indent=12).lstrip()

    # If there are multiple lines in ORGANISM
    if '\n' in ORGANISM:
        lines = ORGANISM.split('\n')
        # The 1st line of ORGANISM should be the species name
        # Starting from the 2nd line, it should be the taxanomy
        # Make everything from the 2nd line a single line without wrapping
        ORGANISM = lines[0] + '\n' + ' ' * 12 + ''.join(lines[1:])

    header = f"""\
LOCUS       {ACCESSION}        {length} bp    {molecule}    {shape}   {division} {get_today()}
DEFINITION  {DEFINITION}
ACCESSION   {ACCESSION}
KEYWORDS    {KEYWORDS}
SOURCE      {SOURCE}
  ORGANISM  {ORGANISM}\n"""

    return header


def init_feature(length, organism, mol_type='genomic DNA') -> str:
    """
    Initialize a feature section of genbank file.
    The feature section always start with the mandatory 'source' feature

    Args:
        length: int
            Length of the genome

        organism: str

        mol_type: str
            Usually 'genomic DNA'
    """
    text = f'''\
FEATURES             Location/Qualifiers
     source          1..{length}
                     /mol_type="{mol_type}"
                     /organism="{organism}"\n'''
    return text


def translate_feature(feature, sequence) -> str:
    """
    Args:
        feature: GenericFeature object

        sequence: str
            DNA sequence to which the feature belongs
    """
    # Build CDS by concatenating introns
    cds = ''
    for start, end, _ in feature.regions:
        cds += sequence[(start - 1):end]

    if feature.strand == '-':
        cds = rev_comp(cds)

    codon_start = feature.get_attribute('codon_start')
    if codon_start is not None:
        cds = cds[(codon_start-1):]

    translation = translate(cds)

    if translation.endswith('*'):
        translation = translation[:-1]  # remove stop codon

    if feature.partial_start and feature.strand == '+':
        return translation  # not full protein --> no need to start with 'M'
    elif feature.partial_end and feature.strand == '-':
        return translation  # not full protein --> no need to start with 'M'
    else:
        return 'M' + translation[1:]  # full protein --> start with 'M'


def format_ref_seq(seq) -> str:
    """
    Take a DNA sequence (str) and format it to the ORIGIN section of genbank file.

    Args:
        seq: str

    Returns: str
    """
    # Remove all ' ' and '\n'
    seq = seq.replace(' ', '').replace('\n', '')

    lines = []
    num_lines = int(len(seq) / 60) + 1  # 60 nucleotides each line
    for i in range(num_lines):
        new = ''  # new line
        pos = i * 60 + 1
        new += ' '*(9 - len(str(pos))) + str(pos)
        for j in range(6):
            a = pos + j * 10 - 1
            b = a + 10
            new += ' ' + seq[a:b]
        lines.append(new)

    return 'ORIGIN\n' + '\n'.join(lines) + '\n'


def write_genbank(
        data,
        file,
        DEFINITION='.',
        KEYWORDS='.',
        SOURCE='.',
        ORGANISM='.',
        genbank_division='ENV',
        use_locus_text=True):
    """
    Write <data> into a new genbank <file>

    Args:
        data: list of Chromosome objects, or dict

            [Chromosome_1, Chromosome_2, ...]

            or

            {
                Chromosome_1.seqname: Chromosome_1,
                Chromosome_2.seqname: Chromosome_2, ...
            }

        file: str, path-like
            The output genbank file

        DEFINITION: str
            The "DEFINITION" field in the genbank file

        KEYWORDS: str
            The "KEYWORDS" field in the genbank file

        SOURCE: str
            The "SOURCE" field in the genbank file

        ORGANISM: str
            The "ORGANISM" field in the genbank file

        genbank_division: str
            The three-letter tag for one of the 18 divisions in the GenBank database
            Default 'ENV' for environmental sampling sequences

        use_locus_text: bool
            Use the genbank_locus_text from the original input gbk file for the LOCUS section
    """
    if type(data) is dict:
        data = data.values()

    with GenbankTextWriter(file) as writer:
        for chromosome in data:
            sequence = chromosome.sequence

            # --- LOCUS text section --- #
            if use_locus_text:
                locus_text = chromosome.genbank_locus_text
            else:
                locus_text = make_header(
                    molecule='DNA',
                    length=len(sequence),
                    shape=['linear', 'circular'][chromosome.circular],
                    ACCESSION=chromosome.seqname,
                    DEFINITION=DEFINITION,
                    KEYWORDS=KEYWORDS,
                    SOURCE=SOURCE,
                    ORGANISM=ORGANISM,
                    division=genbank_division
                )

            # --- FEATURES text section --- #
            features_text = init_feature(
                length=len(sequence),
                organism=ORGANISM.split('\n')[0]
            )

            for f in chromosome.feature_array:
                # 'source' feature was made in the _init_feature()
                if f.type == 'source':
                    continue

                # For CDS, make 'translation', 'codon_start' if absent
                if f.type == 'CDS':
                    if f.get_attribute('translation') is None:
                        aa = translate_feature(feature=f, sequence=sequence)
                        f.add_attribute('translation', aa)
                    if f.get_attribute('codon_start') is None:
                        f.set_attribute('codon_start', f.frame)

                features_text += generic_feature_to_genbank_text(feature=f)

            # --- ORIGIN text sections --- #
            origin_text = format_ref_seq(sequence)

            # --- Write into file --- #
            writer.write(
                GenbankText(
                    locus_text=locus_text,
                    features_text=features_text,
                    origin_text=origin_text
                )
            )


def make_genbank(
        fasta,
        gtf,
        output,
        shape='linear',
        DEFINITION='.',
        KEYWORDS='.',
        SOURCE='.',
        ORGANISM='.',
        genbank_division='ENV'):
    """
    Merge a fasta file and a GTF file into a genbank file, in which
        the fasta file provides the genome sequence and
        the GTF file provides annotation (e.g. CDS).

    The fasta file can contain more than one sequence. Headers in the fasta file
        is used as the LOCUS name (i.e. ACCESSION) in the genbank file.

    GFF and GTF formats are both supported.

    Args:
        fasta: str, path-like
            The input fasta file

        gtf: str, path-like
            The input GTF (or GFF) file

        output: str, path-like
            The output genbank file

        shape: str
            The shape of DNA or RNA, either 'linear' or 'circular'

        DEFINITION: str
            The "DEFINITION" field in the genbank file

        KEYWORDS: str
            The "KEYWORDS" field in the genbank file

        SOURCE: str
            The "SOURCE" field in the genbank file

        ORGANISM: str
            The "ORGANISM" field in the genbank file

        genbank_division: str,
            The three-letter tag for one of the 18 divisions in the GenBank database.
            Default 'ENV' for environmental sampling sequences
    """
    data = read_fasta_gtf(
        fasta=fasta,
        gtf=gtf,
        circular=(shape == 'circular')
    )

    write_genbank(
        data=data,
        file=output,
        DEFINITION=DEFINITION,
        KEYWORDS=KEYWORDS,
        SOURCE=SOURCE,
        ORGANISM=ORGANISM,
        genbank_division=genbank_division
    )



from typing import Optional, Tuple, Union, Dict, List


class FastaParser:

    def __init__(self, file: str):
        self.__fasta = open(file, 'r')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return

    def __iter__(self):
        self.__fasta.seek(0)
        return self

    def __next__(self):
        r = self.next()
        if r:
            return r
        else:  # r is None
            raise StopIteration

    def next(self) -> Optional[Tuple[str, str]]:
        """
        Returns the next read of the fasta file
        If it reaches the end of the file, return None
        """
        header = self.__fasta.readline().rstrip()[1:]
        if header == '':
            return None

        seq = ''
        while True:
            pos = self.__fasta.tell()
            line = self.__fasta.readline().rstrip()
            if line.startswith('>'):
                self.__fasta.seek(pos)
                return header, seq
            if line == '':
                return header, seq
            seq = seq + line

    def close(self):
        self.__fasta.close()


class FastaWriter:

    def __init__(self, file: str, mode: str = 'w'):
        """
        Args:
            file: path-like

            mode: 'w' for write or 'a' for append
        """
        assert mode in ['w', 'a']

        self.__fasta = open(file, mode)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __wrap(self, seq: str, length: int) -> str:
        """
        Wraps a single line of string by a specified length

        Args:
            seq: a single line of DNA or protein sequence without '\n'

            length: length of each line
        """
        if len(seq) <= length:
            return seq  # no need to wrap

        w = length
        list_ = []
        for i in range(int(len(seq)/w) + 1):
            list_.append(seq[i*w:(i+1)*w])

        return '\n'.join(list_)

    def write(self, header: str, sequence: str, wrap: int = 80):
        """
        Args:
            header: Fasta header

            sequence: DNA or protein sequence without '\n'

            wrap: length of each wrapped line for the sequence
        """
        seq = self.__wrap(sequence, wrap)
        self.__fasta.write(f'>{header}\n{seq}\n')

    def close(self):
        self.__fasta.close()


def read_fasta(
        file: str,
        as_dict: bool = False,
        strip_header: bool = False) \
        -> Union[Dict[str, str], List[Tuple[str, str]]]:
    """
    Args:
        file: path-like

        as_dict:
            Return dictionary or not

        strip_header:
            Strip everything after the first space in the header

    Returns:
        {
            header: sequence, ...
            header: sequence, ...
        }

        [
            (header, sequence), (header, sequence), ...
        ]
    """

    parser = FastaParser(file)

    if as_dict:
        if strip_header:
            data = {h.split(' ')[0]: s for h, s in parser}
        else:
            data = {h: s for h, s in parser}
    else:
        if strip_header:
            data = [(h.split(' ')[0], s) for h, s in parser]
        else:
            data = [(h, s) for h, s in parser]

    parser.close()

    return data


def write_fasta(
        data: Union[Dict[str, str], List[Tuple[str, str]]],
        file: str):
    """
    Args:
        data:
            Fasta data return by the function 'read_fasta'

        file: path-like
    """

    if type(data) is list:
        iterator = data
    else:
        iterator = data.items()

    with FastaWriter(file) as writer:
        for head, seq in iterator:
            writer.write(header=head, sequence=seq)


def subset_fasta(file, headers, output):
    """
    Args:
        file: str, path-like
            The input fasta file

        headers: str, or list of str
            Each str is a header to be included

        output: str, path-like
            The output fasta file
    """
    if isinstance(headers, str):
        headers = [headers]

    with FastaParser(file) as parser:
        with FastaWriter(output) as writer:
            for head, seq in parser:
                if head in headers:
                    writer.write(head, seq)



from typing import List, Union, Optional, Tuple, Dict
from .lowlevel import call, printf
from .file_conversion import sam_to_bam


def sort_bam(file: str, keep: bool = False):
    """
    Args:
        file: path-like

        keep:
            Keep the input file or not
    """
    file_out = f"{file[:-4]}_sorted.{file[-3:]}"
    call(f"samtools sort {file} > {file_out}")
    if not keep:
        call(f"rm {file}")
        call(f"mv {file_out} {file}")


def index_bam(file: str):
    """
    Args:
        file: path-like
    """
    cmd = f"samtools index {file}"
    call(cmd)


def sam_to_indexed_bam(file: str, keep: bool = True):
    """
    Args:
        file: path-like
            The input sam file

        keep:
            Keep the input file or not
    """
    sam_to_bam(file, keep=keep)
    bam = file[:-4] + '.bam'
    sort_bam(bam, keep=False)
    index_bam(bam)


def subset_bam_regions(
        file: str,
        regions: List[str],
        output: Optional[str] = None,
        keep: bool = True):
    """
    Args:
        file: path-like
            The input bam file

        regions:
            Each str is a region of the reference genome, e.g.
                chr1            chromosome 1
                chr3:1000-2000  chromosome 3 from 1000th (inclusive) to 2000th (inclusive) base

        output: path-like
            The output bam file
            If None, add subscript '_subset' to the input bam file

        keep:
            If False, delete the input <file> and rename the output as the input <file>
            Overrides the <output> file name
    """
    # Convert regions list into a string
    # ['chr1', 'chr2:1001-2000'] -> ' chr1 chr2:1001-2000'
    regions = ' ' + ' '.join(regions)  # space-separated regions

    if output is None:
        output = file[:-len('.bam')] + '_subset.bam'

    # -b: output is a bam
    # -h: include header section
    call(f"samtools view -b -h {file}{regions} > {output}")

    if not keep:
        call(f"rm {file}")
        call(f"mv {output} {file}")


def remove_unmapped(file: str, keep: bool = True):
    """
    Remove unmapped reads from an input SAM/BAM file
        by using the option '-F 4' of 'samtools view' command

    Args:
        file: path-like
            The input sam/bam file

        keep:
            Keep the input file or not
    """
    b = ['', '-b '][file.endswith('.bam')]  # Output file is bam or not
    file_out = f"{file[:-4]}_remove_unmapped.{file[-3:]}"

    # -b: output is a bam
    # -h: include header section
    # -F 4: NOT including the flag 'read unmapped'
    call(f"samtools view -h -F 4 {b}{file} > {file_out}")

    if not keep:
        call(f"rm {file}")
        call(f"mv {file_out} {file}")


class SamParser:
    """
    A SAM file parser
    """
    def __init__(self, file: str):
        """
        The header section is parsed upon instantiation.

        Args:
            file: path-like
        """
        self.__sam = open(file, 'r')
        header = ''
        while True:
            # Get the current position
            pos = self.__sam.tell()
            # Readline and move on to the next position
            line = self.__sam.readline()
            if line.startswith('@'):
                header = header + line
            else:
                # If reaching the alignment section, that is,
                #   the header section has been parsed completely,
                #   then go back one line and break out the loop
                self.__sam.seek(pos)
                # Remember where the first line of data is
                self.pos_0 = pos
                break
        # Store the header string
        self.header = header

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return

    def __iter__(self):
        self.__sam.seek(self.pos_0)
        return self

    def __next__(self):
        r = self.next()
        if r:
            return r
        else:  # r is None
            raise StopIteration

    def next(self) -> Optional[Tuple[Union[str, int]]]:
        """
        Each line of the SAM file has at least 11 fields

        #   Col	Field   Type    Description
        0   1   QNAME   String  Query template NAME
        1   2   FLAG    Int     bitwise FLAG
        2   3   RNAME   String  References sequence NAME
        3   4   POS     Int     1-based leftmost mapping POSition
        4   5   MAPQ    Int     MAPping Quality
        5   6   CIGAR   String  CIGAR String
        6   7   RNEXT   String  Ref. name of the mate/NEXT read
        7   8   PNEXT   Int     Position of the mate/NEXT read
        8   9   TLEN    Int     observed Template LENgth
        9   10  SEQ     String  segment SEQuence
        10  11  QUAL    String  ASCII of Phred-scaled base QUALity+33
        11  12  optional fields...

        Returns:
            The length of tuple might be > 11,
              depending of the presence of additional optional fields
        """
        line = self.__sam.readline().rstrip()
        if line:
            fields: List[Union[str, int]] = line.split('\t')
            for i in (1, 3, 4, 7, 8):
                fields[i] = int(fields[i])
            return tuple(fields)
        else:  # line == ''
            return None

    def close(self):
        self.__sam.close()


class SamWriter:
    def __init__(self, file: str, header: str, mode: str = 'w'):
        """
        The header section should be written in upon instantiation.

        Args:
            file: path-like

            header:
                The header section

            mode:
                The file mode: 'w' or 'a'
        """
        self.__sam = open(file, mode)
        if not header == '':
            self.__sam.write(header.rstrip() + '\n')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return

    def write(self, read: Tuple[Union[str, int]]):
        """
        Args:
            read:
                Containing 11 (at least) fields of a line of SAM file
        """
        self.__sam.write('\t'.join(map(str, read)) + '\n')

    def close(self):
        self.__sam.close()


FLAGS_PROPERTIES = (
    'read paired',  # 0
    'read mapped in proper pair',  # 1
    'read unmapped',  # 2
    'mate unmapped',  # 3
    'read reverse strand',  # 4
    'mate reverse strand',  # 5
    'first in pair',  # 6
    'second in pair',  # 7
    'not primary alignment',  # 8
    'read fails platform/vendor quality checks',  # 9
    'read is PCR or optical duplicate',  # 10
    'supplementary alignment'  # 11
)


def decode_flag(flag: Union[int, str], return_tuple: bool = False):
    """
    The flag in SAM files is a decimal integer of 12 bits for 12 properties.
    The range of the integer is 0 - 4095 (2^12 -1)
    This function decodes the flag and returns an easily readable format.

    000000000000
    ||||||||||||
    |||||||||||read paired
    ||||||||||read mapped in proper pair
    |||||||||read unmapped
    ||||||||mate unmapped
    |||||||read reverse strand
    ||||||mate reverse strand
    |||||first in pair
    ||||second in pair
    |||not primary alignment
    ||read fails platform/vendor quality checks
    |read is PCR or optical duplicate
    supplementary alignment

    Args:
        flag:
            decimal flag, ranging from 0 - 4095

        return_tuple:
            If True, returns a tuple of True/False, to speed up the function

    Returns:
        A dictionary of the properties of the flag, with
            each key as the property name
            each value as True/False
    """
    # Format decimal to binary strings
    bin_str = '{0:b}'.format(int(flag))
    # Pad zeros up to 12 bits
    bin_str = '0'*(12-len(bin_str)) + bin_str
    # Starting from the right-most bit, make a list of True/False
    bin_tuple = tuple(bool(int(b)) for b in bin_str[::-1])
    if return_tuple:
        return bin_tuple
    # Make {PROPERTY: True/False} pairs and return the dictionary
    return {key: val for key, val in zip(FLAGS_PROPERTIES, bin_tuple)}


def encode_flag(flag_dict: Dict[str, bool], return_str: bool = False):
    """
    Encode a dictionary of flags into an int, which represents the 12-bit binary string

    Args:
        flag_dict:
            For example:

            {'first in pair': True,
             'read reverse strand': False},

            Flags not specified are defaulted False

        return_str: bool
            If True, returns a 12-bit str

    Returns: int or str
    """

    bits = ['0'] * 12
    for key, val in flag_dict.items():
        # This flag is True, set the corresponding bit to '1'
        if val:
            which = FLAGS_PROPERTIES.index(key)
            bits[which] = '1'
    # First flag is on the right (last element), so reverse the list <bits>
    bits = bits[::-1]
    bits = ''.join(bits)
    if return_str:
        return bits
    return int(bits, 2)  # Binary string -> int


def filter_sam_by_flag(
        file_in: str,
        file_out: str,
        flag_sets: List[Dict[str, bool]]):
    """
    Args:
        file_in: path-like
            The input SAM file

        file_out: path-like
            The output SAM file

        flag_sets:
            Each dictionary is a set of flags that have to be completely satisfied.
            There could be more than one set of flags, therefore many dictionaries.
            If a read satisfies any one of the dictionary, then it is included, so other
                there's no need to look into other dictionaries in the list.

            For example:

            [
                {'first in pair': True,
                 'read reverse strand': False},

                {'first in pair': False,
                 'read reverse strand': True},
            ]

            The two dictionaries represent two sets of flags:
                "read 1 mapped to the forward strand"
                "read 2 mapped to the reverse strand"

            These two sets of flags indicate the DNA insert is from the forward strand.
    """
    parser = SamParser(file_in)
    writer = SamWriter(file_out, header=parser.header)

    for A in parser:
        decoded_flag = decode_flag(A[1])  # FLAG is the second field of each line
        for each_set in flag_sets:
            # each_set is a dict

            passed = True
            for key in each_set.keys():
                if each_set[key] != decoded_flag[key]:
                    # If any flag among the set is not satisfied,
                    #   break from the loop of the dictionary <each_set>
                    passed = False
                    break

            if passed:
                # If all of the flags among the set is satisfied,
                #   write the read into the output sam file
                writer.write(A)
                # Break from the loop of all sets,
                # A read can only be included once, even if it satisfies other set of flags.
                break

    parser.close()
    writer.close()


def print_flag(flag: Optional[Union[int, Dict[str, bool]]] = None):
    """
    Decode (int -> dict) or encode (dict -> int) a flag and then print it

    Args: None or int or dict
        If None:
            Print an example dictionary of flags

        If int:
            Print decoded (int -> dict) flag

        If dict:
            print encode (dict -> int) flag
    """
    if type(flag) == dict:
        printf(encode_flag(flag))
        return

    elif flag is None:
        D = {key: True for key in FLAGS_PROPERTIES}

    elif type(flag) == int:
        D = decode_flag(flag)

    t = ''
    for key, val in D.items():
        t = t + f"'{key}': {val},\n "
    t = '{' + t[:-3] + '}'
    printf(t)


def print_sam(read: Tuple[Union[str, int]] = None):
    """
    Pretty print a read (tuple) from sam file

    Args:
        read:
            Containing (at least) 11 fields of a read from sam file
    """
    if read is None:
        text = """\
#   Col	Field   Type    Description
0   1   QNAME   String  Query template NAME
1   2   FLAG    Int     bitwise FLAG
2   3   RNAME   String  References sequence NAME
3   4   POS     Int     1-based leftmost mapping POSition
4   5   MAPQ    Int     MAPping Quality
5   6   CIGAR   String  CIGAR String
6   7   RNEXT   String  Ref. name of the mate/NEXT read
7   8   PNEXT   Int     Position of the mate/NEXT read
8   9   TLEN    Int     observed Template LENgth
9   10  SEQ     String  segment SEQuence
10  11  QUAL    String  ASCII of Phred-scaled base QUALity+33"""
        printf(text)

    elif type(read) is tuple or type(read) is list:
        fields = ['QNAME', 'FLAG ', 'RNAME', 'POS  ', 'MAPQ ',
                  'CIGAR', 'RNEXT', 'PNEXT', 'TLEN ', 'SEQ  ', 'QUAL ']
        for i in range(11):
            printf(f"{i}\t{fields[i]}\t{read[i]}")
        if len(read) > 11:
            for i in range(11, len(read)):
                printf(f"{i}\t     \t{read[i]}")



from .dnatools import rev_comp
from .genbank_parse import read_genbank
from .genbank_write import write_genbank
from .gtftools import read_gtf, write_gtf
from .fasta import read_fasta, write_fasta
from .dataclass import FeatureArray, Chromosome


def _find_feature(feature_arr, keywords):
    """
    In a FeatureArray object, if any one of the GenericFeature contains any one of the <keywords>,
        then return the GenericFeature object

    Args:
        feature_arr: FeatureArray object

        keywords: list of str
    """
    # For each GenericFeature
    for f in feature_arr:
        # For each keyword
        for word in keywords:
            # For each value in the attributes [(key_1, val_1), (key_2, val_2), ...]
            for key, val in f.attributes:
                if word in str(val):
                    return f
    return None


def locus_extractor(fasta, gtf, keywords, flank, fasta_out, gtf_out):
    """
    Use <keywords> to find a 'seed' feature (i.e. CDS) in each contig and
        extract the genomic region from (seed.start - flank) to (seed.end + flank)

    If the seed feature is on the reverse strand,
        reverse the genomic region

    Does not support circular genome, thus if a GtfFeature has start > end position, then
        that feature will be discarded

    Args:
        fasta: str, path-like
            The input fasta file

        gtf: str, path-like
            The input GTF file

        keywords: list of str

        flank: int

        fasta_out: str, path-like
            The output fasta file

        gtf_out: str, path-like
            The output GTF file
    """
    if type(keywords) is str:
        keywords = [keywords]

    fasta_dict = read_fasta(fasta, as_dict=True)
    gtf_dict = read_gtf(gtf, as_dict=True)

    __sequence_dict = dict()
    __feature_dict = dict()

    # Iterate through each contig
    for seqname, sequence in fasta_dict.items():
        gtf_features = gtf_dict.get(seqname, [])
        if len(gtf_features) == 0: continue  # no annotation, skip this contig

        feature_array = FeatureArray(seqname=seqname, genome_size=len(sequence), features=gtf_features, circular=False)

        seed = _find_feature(feature_array, keywords)
        if seed is None: continue  # seed not found, skip this contig

        region_start = seed.start - flank
        region_end = seed.end + flank

        region_start = max(region_start, 1)  # out of bound

        # Extract sequence from fasta
        sequence = sequence[region_start-1: region_end]

        # Crop features from the feature array
        feature_array.crop(region_start, region_end)

        # Reverse sequence and features if seed is on the reverse strand
        if seed.strand == '-':
            sequence = rev_comp(sequence)
            feature_array.reverse()

        __sequence_dict[seqname] = sequence
        __feature_dict[seqname] = feature_array

    # Write output files
    write_fasta(data=__sequence_dict, file=fasta_out)
    write_gtf(data=__feature_dict, file=gtf_out)


def genbank_locus_extractor(genbank, keywords, flank, output):
    """
    Similar to locus_extractor() except the genomic sequence and annotation are
        read from a genbank file

    Args:
        genbank: str, path-like
            The input genbank file

        keywords: str, or list of str

        flank: int

        output: str, path-like
            The output genbank file
    """
    if type(keywords) is str:
        keywords = [keywords]

    loci = []  # list of Chromosome objects

    # Iterate through each contig
    for chromosome in read_genbank(genbank):  # read_genbank() -> list of Chromosome objects
        seqname = chromosome.seqname
        feature_array = chromosome.feature_array
        sequence = chromosome.sequence

        seed = _find_feature(feature_array, keywords)
        if seed is None:
            continue  # seed not found, skip this chromosome

        region_start = seed.start - flank
        region_end = seed.end + flank

        region_start = max(region_start, 1)  # out of bound
        region_end = min(region_end, len(sequence))

        # Add information to the seqname
        seqname = seqname + f';from={region_start};to={region_end};strand={seed.strand}'

        # Extract sequence from fasta
        sequence = sequence[region_start-1: region_end]

        # Crop features from the feature array
        feature_array.crop(region_start, region_end)

        # Reverse sequence and features if seed is on the reverse strand
        if seed.strand == '-':
            sequence = rev_comp(sequence)
            feature_array.reverse()

        loci.append(
            Chromosome(
                seqname=seqname,
                sequence=sequence,
                feature_array=feature_array,
                circular=False  # Must be linear molecule after locus extraction
            )
        )

    # Write output files
    write_genbank(data=loci, file=output)